// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#ifndef _HOBBESCONF_H_
#define _HOBBESCONF_H_

#include "output/outputresultsink.h"
#include <string>
#include <sstream>
#include <limits>

enum SamOutputOptions {
	SAM_DEFAULT, SAM_NO_HEADERS, SAM_NO_SEQ
};

enum OutputFormat {
	TEXT_FORMAT,
	BOWTIE_FORMAT,
	SAM_FORMAT,
	DEBUG,
};

enum OutputOptions {
	WRITE_STATS = 1,
	WRITE_DEBUG_INFO = 2,
	WRITE_MAPPED_READS = 16,
	WRITE_UNMAPPED_READS = 32,
	WRITE_DISCARDED_READS = 64,
	WRITE_CANDIDATES = 128,
	WRITE_READ_DISTRIBUTION = 256
};

enum OutputStreamType {
	STANDARD_OUTPUT, FILE_OUTPUT
};

enum IndexOptions {
	CREATE_INDEX = 1, LOAD_INDEX = 2
};

enum InputStreamType {
	COMMAND_LINE_INPUT, STANDARD_INPUT, FILE_INPUT, INVALID_INPUT_STREAM
};

enum InputCompression {
	UNCOMPRESSED,
	GZIP,
	BZIP2
};

enum InputFormat {
  READS_TXT_FORMAT, READS_FASTQ_FORMAT, READS_PAIRED_FASTQ_FORMAT, READS_FASTA_FORMAT, INVALID_INPUT_FORMAT
};

enum ConfOptions {
	TEST_BRUTE_FORCE = 1,
	DISTRIBUTION_READS = 2,
	GENERATE_ONLY_SUBSTITUTION_READS = 4,
	ALL_MAPPINGS = 256,
	BEST_MAPPINGS = 512,
	ONLY_FORWARD_ALIGN = 1024,
	ONLY_REVERSE_ALIGN = 2048,
};

enum ReadType {
	SINGLE,
	PAIRED,
	PAIRED_ENHANCED,
};

enum QueryType {
	SELECT_SUBST,
	SELECT_INDEL,
	SELECT_HEURISTIC_INDEL,
	INVALID_QUERYTYPE
};

class HobbesConf {
public:
	// Input.
	// Reference sequence.
	std::string refSeqPath;
	bool refSeqPathIsDir;
	// Reads.
	std::string readsFilename;
	std::string secondReadsFilename;
	std::string indexFileName;
	InputStreamType inputStreamType;
	InputCompression inputCompression;
	InputFormat inputFormat;
	ReadType readType;
	unsigned inputMinOutstanding;
	unsigned inputMaxOutstanding;
	// Reading command line read string into this.
	std::stringstream inputReadStream;

	// Output.
	OutputStreamType outputStreamType;
	OutputFormat outputFormat;
	unsigned outputOptions;
	SamOutputOptions samOutputOptions;
	OutputResultSinkMode outputResultSinkMode;
	unsigned outputMaxOutstanding;
	std::string mappedOutputFileName;
	std::string unmappedOutputFileName;
	bool genCigar;
	bool useProgressIndicator;

	// General options.
	unsigned confOptions;
	unsigned indexOptions;
	unsigned gramLength;

	// Mapping options.
	QueryType queryType;
	unsigned distThresh;
	unsigned numReads;
	unsigned maxMappingsPerRead;
	unsigned numThreads;

	// For paired-end mapping,
	unsigned minInsertSize;
	unsigned maxInsertSize;

	std::string commandLineInput;
	std::string hobbesVersion;

	HobbesConf() :
		refSeqPathIsDir(false), inputStreamType(INVALID_INPUT_STREAM),
		inputCompression(UNCOMPRESSED), inputFormat(INVALID_INPUT_FORMAT), readType(SINGLE),
		inputMinOutstanding(100), inputMaxOutstanding(10000),
		outputStreamType(STANDARD_OUTPUT), outputFormat(SAM_FORMAT), outputOptions(0),
		samOutputOptions(SAM_DEFAULT), outputResultSinkMode(ORSM_ASYNC_FIFO),
		outputMaxOutstanding(100000), genCigar(false), useProgressIndicator(true),
		confOptions(0), indexOptions(0), gramLength(0), queryType(INVALID_QUERYTYPE),
		distThresh(std::numeric_limits<unsigned>::max()),
		numReads(std::numeric_limits<unsigned>::max()),
		maxMappingsPerRead(std::numeric_limits<unsigned>::max()), numThreads(1),
		minInsertSize(std::numeric_limits<unsigned>::max()),
		maxInsertSize(std::numeric_limits<unsigned>::max()) {
		inputReadStream.str(std::string());
	}

	// For checking sanity of configuration.
	void checkMappingConf();
	void checkIndexConstructionConf();
};

# endif
