// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#ifndef _INDEXER_H_
#define _INDEXER_H_

#include "gramgen.h"
#include "hobbesconf.h"
#include "hobbesstats.h"
#include "bitvectorfilter.h"

#include <limits>
#include <string>
#include <tr1/unordered_map>

typedef std::tr1::unordered_map<unsigned, std::vector<BitVecPair>*> BitVectorInvertedIndex;
typedef std::pair<unsigned, std::vector<BitVecPair>*>  BitVectorInvertedListWithOffset;

class Indexer {
public:
	std::string* refSeq;
	GramGen gramGen;
	BitVectorInvertedIndex bitVecRefInvIndex;

	// Chromosomes table.
	std::vector<std::string> chromosomeNames;
	std::vector<unsigned> chromosomeLengths;

	Indexer(HobbesConf& conf, HobbesStats& stats) :
		refSeq(new std::string()), conf(conf), stats(stats),
		refSeqPath(&conf.refSeqPath), indexFileName(new std::string(conf.indexFileName)) {
	}

	void readRefSeqFromDir();
	void readRefSeqFromFile(const std::string& refSeqFile);
	void buildIndex();
	void loadIndex();

	~Indexer() {
		if (refSeq != NULL) {
			delete refSeq;
		}
		if (indexFileName != NULL) {
			delete indexFileName;
		}
		for (BitVectorInvertedIndex::iterator itr = bitVecRefInvIndex.begin();
				itr != bitVecRefInvIndex.end(); itr++) {
			delete itr->second;
		}
	}

private:
	HobbesConf& conf;
	HobbesStats& stats;
	void setGramGen() {
		gramGen.initialize(conf.gramLength);
	}

	std::string* refSeqPath;
	std::string* indexFileName;

	void addGramToIndex(unsigned hash, unsigned pos);
	void writeIndexToFile();
	void setIndexStats();
};

class BitVecTriple {
public:
	unsigned hash;
	BitVecPair bitVecPair;

	BitVecTriple(unsigned hash, BitVecPair bitVecPair) :
		hash(hash), bitVecPair(bitVecPair) {
	}

	// Initialize with high value to appear at end when sorting.
	BitVecTriple() :
		hash(std::numeric_limits<unsigned>::max()) {
	}

	static bool cmp(const BitVecTriple& a, const BitVecTriple& b) {
		if (a.hash == b.hash) {
			return a.bitVecPair.pos < b.bitVecPair.pos;
		}
		return a.hash < b.hash;
	}
};

#endif
