// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#ifndef _PAIREDENDVERIFIER_H_
#define _PAIREDENDVERIFIER_H_

#include "search/read.h"
#include "search/pairedendverifier.h"
#include "search/singleendverifier.h"

#include <sstream>
#include <string>
#include <vector>

class PairedReadMapping: public ReadMapping {
public:
	unsigned posRight;
	unsigned distanceRight;
	const std::stringstream* cigarRight;

	PairedReadMapping(unsigned posLeft, unsigned distanceLeft, unsigned posRight,
			unsigned distanceRight, const std::stringstream* cigarLeft,
			const std::stringstream* cigarRight) :
				ReadMapping(posLeft, distanceLeft, cigarLeft), posRight(posRight),
				distanceRight(distanceRight), cigarRight(cigarRight) {
	}

	virtual void clear() {
		ReadMapping::clear();
		if (cigarRight != NULL) {
			delete cigarRight;
		}
	}
};

class PairedEndVerifier: public Verifier {
public:
	SingleEndVerifier* singleEndVerifier;

	PairedEndVerifier(Indexer* indexer, unsigned distThresh, unsigned minInsertSize,
			unsigned maxInsertSize, SingleEndVerifier* singleEndVerifier) :
				Verifier(indexer, distThresh), singleEndVerifier(singleEndVerifier),
				minInsertSize(minInsertSize), maxInsertSize(maxInsertSize) {
	}

	void setMinMaxDist(unsigned readLen);

	virtual void verifyCandidates(PairedEndRead& read, std::vector<unsigned>& candidates1,
			std::vector<unsigned>& candidates2, std::vector<PairedReadMapping>& results);

	virtual ~PairedEndVerifier() {
		delete singleEndVerifier;
	}

protected:
	unsigned minInsertSize;
	unsigned maxInsertSize;
	unsigned minDist;
	unsigned maxDist;
};

class PairedEndVerifierEnhanced: public PairedEndVerifier {
public:
	PairedEndVerifierEnhanced(Indexer* indexer, unsigned distThresh, unsigned minInsertSize,
			unsigned maxInsertSize, SingleEndVerifier* singleEndVerifier) :
			PairedEndVerifier(indexer, distThresh, minInsertSize, maxInsertSize, singleEndVerifier) {
	}

	void verifyCandidates(PairedEndRead& read, std::vector<unsigned>& candidates1,
			std::vector<unsigned>& candidates2, std::vector<PairedReadMapping>& results,
			std::vector<ReadMapping>& results1, std::vector<ReadMapping>& results2);
};

class PairedEndHammingVerifier: public PairedEndVerifier {
public:
	PairedEndHammingVerifier(Indexer* indexer, unsigned distThresh, unsigned minInsertSize,
			unsigned maxInsertSize) :
				PairedEndVerifier(indexer, distThresh, minInsertSize, maxInsertSize,
						new SingleEndHammingVerifier(indexer, distThresh)) {
	}
};

class PairedEndIndelVerifier: public PairedEndVerifier {
public:
	PairedEndIndelVerifier(Indexer* indexer, unsigned distThresh, bool genCigar,
			unsigned minInsertSize, unsigned maxInsertSize) :
				PairedEndVerifier(indexer, distThresh, minInsertSize, maxInsertSize,
						new SingleEndIndelVerifier(indexer, distThresh, genCigar)),
				genCigar(genCigar) {
	}

	void verifyCandidates(PairedEndRead& read, std::vector<unsigned>& candidates1,
			std::vector<unsigned>& candidates2, std::vector<PairedReadMapping>& results);

protected:
	bool genCigar;
};

class PairedEndHammingVerifierEnhanced: public PairedEndVerifierEnhanced {
public:
	PairedEndHammingVerifierEnhanced(Indexer* indexer, unsigned distThresh, unsigned minInsertSize,
			unsigned maxInsertSize) :
				PairedEndVerifierEnhanced(indexer, distThresh, minInsertSize, maxInsertSize,
						new SingleEndHammingVerifier(indexer, distThresh)) {
	}
};

class PairedEndIndelVerifierEnhanced: public PairedEndVerifierEnhanced {
public:
	PairedEndIndelVerifierEnhanced(Indexer* indexer, unsigned distThresh, bool genCigar,
			unsigned minInsertSize, unsigned maxInsertSize) :
				PairedEndVerifierEnhanced(indexer, distThresh, minInsertSize, maxInsertSize,
						new SingleEndIndelVerifier(indexer, distThresh, genCigar)) {
	}
};

class PairedEndHeuristicIndelVerifierEnhanced: public PairedEndVerifierEnhanced {
public:
	PairedEndHeuristicIndelVerifierEnhanced(Indexer* indexer, unsigned distThresh, bool genCigar,
			unsigned minInsertSize, unsigned maxInsertSize) :
				PairedEndVerifierEnhanced(indexer, distThresh, minInsertSize, maxInsertSize,
						new SingleEndHeuristicIndelVerifier(indexer, distThresh, genCigar)) {
	}
};

#endif
