// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "bitvectorfilter.h"
#include <limits>
#include <string>

using namespace std;

BitvectorType generateBitVector(BitvectorType& alignMask, const string& seq, unsigned startPos,
		unsigned gramLen, unsigned seqLen) {
	BitvectorType bitVector = 0;
	unsigned step = (sizeof(BitvectorType) * 8) / 2;
	bool partialLeft = false;
	bool partialRight = false;
	unsigned loopStart = 0;
	unsigned loopStop = 0;

	// Set loopStart and loopStop, and detect partialLeft and/or partialRight.
	if (startPos >= step) {
		if ((startPos + gramLen) <= (seqLen - step)) {
			// Full left and right.
			loopStart = startPos - step;
			loopStop = startPos + gramLen + step - 1;
		} else {
			// Full left + no right or partial right.
			loopStart = startPos - step;
			partialRight = true;
			// Differentiating between partial right and no right.
			if ((startPos + gramLen) >= seqLen) {
				loopStop = startPos - 1;
			} else {
				loopStop = seqLen - 1;
			}
		}
	} else {
		if ((startPos + gramLen) <= (seqLen - step)) {
			// No left or partial left + full right.
			loopStart = 0;
			loopStop = startPos + gramLen + step - 1;
			partialLeft = true;
		} else {
			// No left or partial left + partial right or no right.
			loopStart = 0;
			partialLeft = true;
			partialRight = true;
			// Differentiating between partial right and no right.
			if ((startPos + gramLen) >= seqLen) {
				loopStop = startPos - 1;
			} else {
				loopStop = seqLen - 1;
			}
		}
	}

	unsigned ptr = loopStart;
	if (ptr == startPos) {
		ptr += gramLen;
	}
	while (ptr < loopStop) {
		switch (seq[ptr]) {
		case BASE_G:
		case BASE_C:
			bitVector = (bitVector + 1) << 1;
			break;
		default:
			bitVector <<= 1;
			break;
		}
		ptr++;
		if (ptr == startPos) {
			ptr += gramLen;
		}
	}

	switch (seq[loopStop]) {
	case BASE_G:
	case BASE_C:
		bitVector++;
		break;
	}

	// Altering bitvector and alignMask for partial right and/or left.
	if (partialLeft && partialRight) {
		bitVector <<= (step - (seqLen - (startPos + gramLen)));
		BitvectorType temp = numeric_limits<BitvectorType>::max();
		// For partial left.
		temp >>= (step - startPos);
		// For partial right.
		alignMask <<= (step - (seqLen - (startPos + gramLen)));
		alignMask &= temp;
	} else if (!partialLeft && partialRight) {
		bitVector <<= (step - (seqLen - (startPos + gramLen)));
		alignMask <<= (step - (seqLen - (startPos + gramLen)));
	} else if (partialLeft && !partialRight) {
		alignMask >>= (step - startPos);
	}
	return bitVector;
}
