// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "hobbesrunner.h"
#include "hobbesconf.h"
#include "hobbesstats.h"
#include "hobbesuserhelp.h"

#include <iostream>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <getopt.h>

using namespace std;

const char* optString = "-Batk:q:p:u:v:f:i:r:c:";

static const struct option longOpts[] = {
		{ "Bowtie-format", no_argument, NULL, 'B' },
		{ "all mappings", no_argument, NULL, 'a' },
		{ "pe", no_argument, NULL, 0 },
		{ "pe2", no_argument, NULL, 0 },
		{ "k mappings per read", required_argument, NULL, 'k' },
		{ "hamming", no_argument, NULL, 0 },
		{ "indel", no_argument, NULL, 0 },
		{ "min", required_argument, NULL, 0 },
		{ "max", required_argument, NULL, 0 },
		{ "allindel", no_argument, NULL, 0 },
		{ "reads fastq file", required_argument, NULL, 'q' },
		{ "seqfq1", required_argument, NULL, 0 },
		{ "seqfq2", required_argument, NULL, 0 },
		{ "pthreads", required_argument, NULL, 'p' },
		{ "num reads", required_argument, NULL, 'u' },
		{ "substitution distance", required_argument, NULL, 'v' },
		{ "reads fasta file", required_argument, NULL, 'f' },
		{ "index file", required_argument, NULL, 'i' },
		{ "reads text file", required_argument, NULL, 'r' },
		{ "read sequence", required_argument, NULL, 'c' },
		{ "nofw", no_argument, NULL, 0 },
		{ "norc", no_argument, NULL, 0 },
		{ "sam-nohead", no_argument, NULL, 0 },
		{ "sam-nosq", no_argument, NULL, 0 },
		{ "sref", required_argument, NULL, 0 },
		{ "dref", required_argument, NULL, 0 },
		{ "mapout", required_argument, NULL, 0 },
		{ "stdin", no_argument, NULL, 0 },
		{ "cigar", no_argument, NULL, 0 },
		{ "gzip", no_argument, NULL, 0 },
		{ "bzip2", no_argument, NULL, 0 },
		{ "noprogress", no_argument, NULL, 0 },
		{ "version", no_argument, NULL, 0 },
		{ "help", no_argument, NULL, 0 },
		{ NULL, no_argument, NULL, 0 } };


int main(int argc, char **argv) {
	// Print help if no args given.
	if (argc == 1) {
		printFile(helpFileName);
		exit(0);
	}

	HobbesRunner runner;

	// Building command line input string.
	runner.conf.commandLineInput.clear();
	for (int i = 0; i < argc; i++) {
		runner.conf.commandLineInput.append(argv[i]);
		runner.conf.commandLineInput.append(" ");
	}

	vector<unsigned> complexOpts;
	int longIndex;
	int opt = getopt_long(argc, argv, optString, longOpts, &longIndex);
	while (opt != -1) {
		switch (opt) {
		case 'B':
			runner.conf.outputFormat = BOWTIE_FORMAT;
			break;
		case 'a':
			runner.conf.confOptions |= ALL_MAPPINGS;
			break;
		case 'k':
			runner.conf.maxMappingsPerRead = atoi(optarg);
			break;
		case 'q':
			runner.conf.inputStreamType = FILE_INPUT;
			runner.conf.inputFormat = READS_FASTQ_FORMAT;
			runner.conf.readsFilename = optarg;
			break;
		case 'f':
			runner.conf.inputStreamType = FILE_INPUT;
			runner.conf.inputFormat = READS_FASTA_FORMAT;
			runner.conf.readsFilename = optarg;
			break;
		case 'c':
			runner.conf.inputStreamType = COMMAND_LINE_INPUT;
			runner.conf.inputFormat = READS_TXT_FORMAT;
			runner.conf.inputReadStream << optarg;
			break;
		case 'r':
			runner.conf.inputStreamType = FILE_INPUT;
			runner.conf.inputFormat = READS_TXT_FORMAT;
			runner.conf.readsFilename = optarg;
			break;
		case 'i':
			runner.conf.indexFileName = optarg;
			break;
		case 'p':
			runner.conf.numThreads = atoi(optarg);
			break;
		case 'u':
			runner.conf.numReads = atoi(optarg);
			break;
		case 'v':
			runner.conf.distThresh = atoi(optarg);
			break;
		case 0:
			if (strcmp("nofw", longOpts[longIndex].name) == 0) {
				runner.conf.confOptions &= ~ONLY_REVERSE_ALIGN;
				runner.conf.confOptions |= ONLY_REVERSE_ALIGN;
			} else if (strcmp("norc", longOpts[longIndex].name) == 0) {
				runner.conf.confOptions &= ~ONLY_FORWARD_ALIGN;
				runner.conf.confOptions |= ONLY_FORWARD_ALIGN;
			} else if (strcmp("sref", longOpts[longIndex].name) == 0) {
				runner.conf.refSeqPathIsDir = false;
				runner.conf.refSeqPath = optarg;
			} else if (strcmp("dref", longOpts[longIndex].name) == 0) {
				runner.conf.refSeqPathIsDir = true;
				runner.conf.refSeqPath = optarg;
			} else if (strcmp("pe", longOpts[longIndex].name) == 0) {
				 // Paired-end uses both forward and reverse only.
				runner.conf.confOptions = ALL_MAPPINGS;
				runner.conf.readType = PAIRED;
			} else if (strcmp("pe2", longOpts[longIndex].name) == 0) {
				 // Paired-end use both forward and reverse only.
				runner.conf.confOptions = ALL_MAPPINGS;
				runner.conf.readType = PAIRED_ENHANCED;
			} else if (strcmp("seqfq1", longOpts[longIndex].name) == 0) {
				runner.conf.inputStreamType = FILE_INPUT;
				runner.conf.inputFormat = READS_PAIRED_FASTQ_FORMAT;
				runner.conf.readsFilename = optarg;
			} else if (strcmp("seqfq2", longOpts[longIndex].name) == 0) {
				runner.conf.secondReadsFilename = optarg;
			} else if (strcmp("hamming", longOpts[longIndex].name) == 0) {
				runner.conf.queryType = SELECT_SUBST;
			} else if (strcmp("indel", longOpts[longIndex].name) == 0) {
				runner.conf.queryType = SELECT_HEURISTIC_INDEL;
			} else if (strcmp("min", longOpts[longIndex].name) == 0) {
				runner.conf.minInsertSize = atoi(optarg);
			} else if (strcmp("max", longOpts[longIndex].name) == 0) {
				runner.conf.maxInsertSize = atoi(optarg);
			} else if (strcmp("allindel", longOpts[longIndex].name) == 0) {
				runner.conf.queryType = SELECT_INDEL;
			} else if (strcmp("sam-nohead", longOpts[longIndex].name) == 0) {
				runner.conf.samOutputOptions = SAM_NO_HEADERS;
			} else if (strcmp("sam-nosq", longOpts[longIndex].name) == 0) {
				runner.conf.samOutputOptions = SAM_NO_SEQ;
			} else if (strcmp("mapout", longOpts[longIndex].name) == 0) {
				runner.conf.outputStreamType = FILE_OUTPUT;
				runner.conf.unmappedOutputFileName = "UNMAP";
				runner.conf.mappedOutputFileName = optarg;
			} else if (strcmp("cigar", longOpts[longIndex].name) == 0) {
				runner.conf.genCigar = true;
			} else if (strcmp("stdin", longOpts[longIndex].name) == 0) {
				runner.conf.inputStreamType = STANDARD_INPUT;
			} else if (strcmp("gzip", longOpts[longIndex].name) == 0) {
				runner.conf.inputCompression = GZIP;
			} else if (strcmp("bzip2", longOpts[longIndex].name) == 0) {
				runner.conf.inputCompression = BZIP2;
			} else if (strcmp("noprogress", longOpts[longIndex].name) == 0) {
				runner.conf.useProgressIndicator = false;
			} else if (strcmp("version", longOpts[longIndex].name) == 0) {
				printFile(versionFileName);
				exit(0);
			} else if (strcmp("help", longOpts[longIndex].name) == 0) {
				printFile(helpFileName);
				exit(0);
			}
			break;
		default: {
			abort();
		}
			break;

		}
		opt = getopt_long(argc, argv, optString, longOpts, &longIndex);
	}

	runner.conf.indexOptions = LOAD_INDEX;
	runner.conf.outputOptions |= WRITE_STATS;
	runner.conf.checkMappingConf();
	runner.run();
	return 0;
}
