// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "hobbesconf.h"
#include <iostream>
#include <cstdlib>

using namespace std;

void HobbesConf::checkMappingConf() {
	// Check input reference sequence and index sanity.
	if (refSeqPath.empty()) {
		cerr << "No input reference sequence specified. Suggestions:" << endl;
		cerr << "--sref <file> = Use reference sequence in given file." << endl;
		cerr << "--dref <dir>  = Use reference sequence in given directory." << endl;
		exit(0);
	}
	if (indexFileName.empty()) {
		cerr << "No Hobbes index specified. Suggestion:" << endl;
		cerr << "-i <file> = Use given Hobbes index." << endl;
		exit(0);
	}
	// Check read input sanity.
	if (inputStreamType == INVALID_INPUT_STREAM || inputFormat == INVALID_INPUT_FORMAT) {
		cerr << "No input reads specified for mapping. Suggestions:" << endl;
		cerr << endl;
		cerr << "Single-end reads:" << endl;
		cerr << "-q <file>   = Get single-end reads from FastQ file." << endl;
		cerr << "-f <file>   = Get single-end reads from FastA file." << endl;
		cerr << "-c <string> = Map given single-end read (only a single read)." << endl;
		cerr << "--stdin     = Get single-end reads from stdin." << endl;
		cerr << endl;
		cerr << "Paired-end reads:" << endl;
		cerr << "--seqfq1 <file> = First FastQ file for paired-end reads." << endl;
		cerr << "--seqfq2 <file> = Second FastQ file for paired-end reads." << endl;
		exit(0);
	}
	if (readType == PAIRED && (readsFilename.empty() || secondReadsFilename.empty())) {
		cerr << "Paired-end mapping requires two input read files. Suggestion:" << endl;
		cerr << "--seqfq1 <file> = First FastQ file for paired-end reads." << endl;
		cerr << "--seqfq2 <file> = Second FastQ file for paired-end reads." << endl;
		exit(0);
	}
	if (readType == PAIRED_ENHANCED && (readsFilename.empty() || secondReadsFilename.empty())) {
		cerr << "Paired-end enhanced mapping requires two input read files. Suggestion:" << endl;
		cerr << "--seqfq1 <file> = First FastQ file for paired-end reads." << endl;
		cerr << "--seqfq2 <file> = Second FastQ file for paired-end reads." << endl;
		exit(0);
	}
	if (maxMappingsPerRead != numeric_limits<unsigned>::max() && (readType == PAIRED || readType == PAIRED_ENHANCED)) {
		cerr << "Hobbes currently does not support finding -k mappings for paired-end reads." << endl;
		cerr << "Only -a is supported for paired-end reads. Sorry." << endl;
		exit(0);
	}
	if (minInsertSize == numeric_limits<unsigned>::max()
			&& (readType == PAIRED || readType == PAIRED_ENHANCED)) {
		cerr << "No minimum insert size given for paired-end mapping. Suggestions:" << endl;
		cerr << "--min <int> = Minimum insert size for paired-end mapping." << endl;
		cerr << "--max <int> = Maximum insert size for paired-end mapping." << endl;
		exit(0);
	}
	if (maxInsertSize == numeric_limits<unsigned>::max()
			&& (readType == PAIRED || readType == PAIRED_ENHANCED)) {
		cerr << "No minimum insert size given for paired-end mapping. Suggestions:" << endl;
		cerr << "--min <int> = Minimum insert size for paired-end mapping." << endl;
		cerr << "--max <int> = Maximum insert size for paired-end mapping." << endl;
		exit(0);
	}
	if (readType == PAIRED_ENHANCED && (readsFilename.empty() || secondReadsFilename.empty())) {
		cerr << "Paired-end enhanced mapping requires two input read files. Suggestion:" << endl;
		cerr << "--seqfq1 <file> = First FastQ file for paired-end reads." << endl;
		cerr << "--seqfq2 <file> = Second FastQ file for paired-end reads." << endl;
		exit(0);
	}
	if (inputStreamType == COMMAND_LINE_INPUT && inputCompression == GZIP) {
		cerr << "Incompatible arguments -c and --gzip."
				<< "Cannot take input from command line arg as gzip.";
		exit(0);
	}
	if (inputStreamType == COMMAND_LINE_INPUT && inputCompression == BZIP2) {
		cerr << "Incompatible arguments -c and --bzip2."
				<< "Cannot take input from command line arg as bzip.";
		exit(0);
	}
	if (inputStreamType == STANDARD_INPUT && inputCompression == GZIP) {
		cerr << "Incompatible arguments --stdin and --gzip."
				<< "Cannot take input from stdin as gzip.";
		exit(0);
	}
	if (inputStreamType == STANDARD_INPUT && inputCompression == BZIP2) {
		cerr << "Incompatible arguments --stdin and --bzip2."
				<< "Cannot take input from stdin as bzip2.";
		exit(0);
	}
	// Check if distance threshold was given.
	if (distThresh == numeric_limits<unsigned>::max()) {
		cerr << "No distance threshold given. Suggestion:" << endl;
		cerr << "-v <int> = Find mappings within given distance threshold." << endl;
		exit(0);
	}
	// Check if distance metric was given.
	if (queryType == INVALID_QUERYTYPE) {
		cerr << "No distance metric given. Suggestion:" << endl;
		cerr << "--hamming = Use Hamming Distance." << endl;
		cerr << "--indel   = Use Edit Distance." << endl;
		cerr << "-v <int> = Find mappings within given distance threshold." << endl;
		exit(0);
	}
	// Check SAM options sanity.
	if (samOutputOptions == SAM_NO_HEADERS && outputFormat != SAM_FORMAT) {
		cerr << "Invalid option --sam-nohead: Please set SAM (-S) format to use this option."
				<< endl;
		exit(0);
	}
	if (samOutputOptions == SAM_NO_SEQ && outputFormat != SAM_FORMAT) {
		cerr << "Invalid option --sam-nosq: Please set SAM (-S) format to use this option." << endl;
		exit(0);
	}
	// Other sanity checks,
	if (maxMappingsPerRead != numeric_limits<unsigned>::max() && (confOptions & ALL_MAPPINGS)) {
		cerr << "Incompatible options -a and -k." << endl;
		cerr << "-a       = Find all mappings per read." << endl;
		cerr << "-k <int> = Find at most given number of mappings per read." << endl;
		exit(0);
	}
	if (indexOptions & CREATE_INDEX) {
		cerr << "Unexpected option CREATE_INDEX set in internal conf." << endl;
		cerr << "This option should only be enabled for the hobbes-index binary." << endl;
		exit(0);
	}
	// Warnings
	if ((minInsertSize != numeric_limits<unsigned>::max()
			|| maxInsertSize != numeric_limits<unsigned>::max()) && readType == SINGLE) {
		cerr << "Warning: Ignoring --min and --max parameters for single-end mapping." << endl;
	}
	// Warnings
	if (numReads == numeric_limits<unsigned>::max() && useProgressIndicator) {
		cerr << "Warning: Specified option -t for progress indicator but -u option not used." << endl;
		cerr <<	"Assuming MAX_INT number of reads to map. Estimated time for completion will be incorrect." << endl;
		cerr << "Suggestion:" << endl;
		cerr << "-t       = Display mapping progress and estimated time to completion." << endl;
		cerr << "-u <int> = Maximum number of reads to map." << endl;
	}
}

void HobbesConf::checkIndexConstructionConf() {
	if (refSeqPath.empty()) {
		cerr << "No input reference sequence specified. Suggestions:" << endl;
		cerr << "--sref <file> = Use reference sequence in given file." << endl;
		cerr << "--dref <file>  = Use reference sequence in given directory." << endl;
		exit(0);
	}
	if (indexFileName.empty()) {
		cerr << "No Hobbes index file specified. Suggestion:" << endl;
		cerr << "-i <file> = Build index into given file." << endl;
		exit(0);
	}
	if (gramLength == 0) {
		cerr << "No gram length specified. We recommend a gram length of 11. Suggestion:" << endl;
		cerr << "-g <int> = Build index with grams of given length." << endl;
		exit(0);
	}
	if (gramLength > 16) {
		cerr << "Gram length > 16 specified." << endl;
		cerr << "Currently, we only support grams up to length 16." << endl;
		cerr << "We recommend a gram length of 11." << endl;
		exit(0);
	}
	if (indexOptions & LOAD_INDEX) {
		cerr << "Unexpected option LOAD_INDEX set in internal conf." << endl;
		cerr << "This option should only be enabled for hobbes binary." << endl;
		exit(0);
	}
}
