// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "input/readinfo.h"
#include "output/resultstringwriter.h"
#include "index/indexer.h"
#include "search/searcher.h"
#include "search/singleendsearcher.h"
#include "search/pairedendsearcher.h"

#include <algorithm>
#include <iostream>
#include <boost/lexical_cast.hpp>

using namespace std;

void NativeFormatResultStringWriter::writeResultString(stringstream& resultStream,
		ReadInfo& readInfo, Searcher* searcher) {
	SingleEndSearcher* concreteSearcher = static_cast<SingleEndSearcher*>(searcher);
	vector<ReadMapping>::iterator forwardIter = concreteSearcher->forwardResults.begin();
	vector<ReadMapping>::iterator forwardEnd = concreteSearcher->forwardResults.end();
	vector<ReadMapping>::iterator reverseIter = concreteSearcher->reverseResults.begin();
	vector<ReadMapping>::iterator reverseEnd = concreteSearcher->reverseResults.end();
	string chrName;
	unsigned relativePos = 0;
	bool entered = false;
	if (forwardIter != forwardEnd) {
		getChromosomeOffset(searcher->indexer->chromosomeNames,
				searcher->indexer->chromosomeLengths, forwardIter->pos, chrName, relativePos);
		resultStream << readInfo.readId << SEPARATOR << FWD_MAPPING << SEPARATOR << chrName
				<< SEPARATOR << relativePos << SEPARATOR;
		resultStream << readInfo.readString << SEPARATOR << readInfo.readQuality << SEPARATOR;
		resultStream << N_TAG << forwardIter->distance << SEPARATOR;
		forwardIter++;
		resultStream << XA_Z_TAG;
		for (; forwardIter != forwardEnd; forwardIter++) {
			entered = true;
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, forwardIter->pos, chrName,
					relativePos);
			resultStream << chrName << COMMA << FWD_MAPPING << relativePos << COMMA
					<< forwardIter->distance << SEMI_COLON;
		}
		for (; reverseIter != reverseEnd; reverseIter++) {
			entered = true;
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, reverseIter->pos, chrName,
					relativePos);
			resultStream << chrName << COMMA << REV_MAPPING << relativePos << COMMA
					<< reverseIter->distance << SEMI_COLON;
		}
	} else if (reverseIter != reverseEnd) {
		// Write the first result.
		getChromosomeOffset(searcher->indexer->chromosomeNames,
				searcher->indexer->chromosomeLengths, reverseIter->pos, chrName, relativePos);
		resultStream << readInfo.readId << SEPARATOR << REV_MAPPING << SEPARATOR << chrName
				<< SEPARATOR << relativePos << SEPARATOR;
		resultStream << readInfo.reverseReadString << SEPARATOR << readInfo.reverseReadQuality
				<< SEPARATOR;
		resultStream << N_TAG << reverseIter->distance << SEPARATOR;
		reverseIter++;
		resultStream << XA_Z_TAG;
		for (; reverseIter != reverseEnd; reverseIter++) {
			entered = true;
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, reverseIter->pos, chrName,
					relativePos);
			resultStream << chrName << COMMA << REV_MAPPING << relativePos << COMMA
					<< reverseIter->distance << SEMI_COLON;
		}
	}
	if (!entered) {
		resultStream << UNAVAILABLE_1 << endl;
	} else {
		resultStream << endl;
	}
}

void PairedNativeFormatResultStringWriter::writeResultString(stringstream& resultStream,
		ReadInfo& readInfo, Searcher* searcher) {
	PairedEndSearcher* concreteSearcher = static_cast<PairedEndSearcher*>(searcher);
	PairedEndReadInfo* concreteReadSt = static_cast<PairedEndReadInfo*>(&readInfo);
	vector<PairedReadMapping>::iterator forwardIter = concreteSearcher->forwardResults.begin();
	vector<PairedReadMapping>::iterator forwardEnd = concreteSearcher->forwardResults.end();
	vector<PairedReadMapping>::iterator reverseIter = concreteSearcher->reverseResults.begin();
	vector<PairedReadMapping>::iterator reverseEnd = concreteSearcher->reverseResults.end();
	string chrName;
	unsigned relativePos = 0;
	unsigned relativePos2 = 0;
	unsigned chrPos = 0;
	bool entered = false;

	resultStream << readInfo.readId << SEPARATOR << FWD_MAPPING << SEPARATOR << chrName << SEPARATOR
			<< relativePos << SEPARATOR;
	resultStream << readInfo.readString << SEPARATOR << readInfo.readQuality << SEPARATOR;
	resultStream << N_TAG << forwardIter->distance << SEPARATOR;

	if (forwardIter != forwardEnd) {
		// Write the first result.
		entered = true;
		getChromosomeOffset(searcher->indexer->chromosomeNames,
				searcher->indexer->chromosomeLengths, forwardIter->pos, chrName, relativePos);
		// For 1-based offset.
		getChromosomeOffset(searcher->indexer->chromosomeNames,
				searcher->indexer->chromosomeLengths, forwardIter->posRight, chrName, relativePos2);
		relativePos++;
		relativePos2++;
		resultStream << concreteReadSt->first.readId << SEPARATOR << FWD_MAPPING << SEPARATOR
				<< chrName << SEPARATOR << relativePos << SEPARATOR;
		resultStream << concreteReadSt->first.readString << SEPARATOR
				<< concreteReadSt->first.readQuality << SEPARATOR;
		resultStream << N_TAG << forwardIter->distance << SEPARATOR;
		forwardIter++;
		vector<PairedReadMapping>::iterator forwardIter2 = forwardIter;
		vector<PairedReadMapping>::iterator reverseIter2 = reverseIter;
		if (forwardIter != forwardEnd) {
			resultStream << XA_Z_TAG;
		}

		for (; forwardIter != forwardEnd; forwardIter++) {
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, forwardIter->pos, chrName, chrPos);
			chrPos++;
			resultStream << chrName << COMMA << FWD_MAPPING << chrPos << COMMA
					<< concreteReadSt->first.readString.size() << CIGAR << COMMA
					<< forwardIter->distance << SEMI_COLON;
		}
		for (; reverseIter != reverseEnd; reverseIter++) {
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, reverseIter->pos, chrName, chrPos);
			chrPos++;
			resultStream << chrName << COMMA << REV_MAPPING << chrPos << COMMA
					<< concreteReadSt->first.reverseReadString.size() << CIGAR << COMMA
					<< reverseIter->distance << SEMI_COLON;
		}
		resultStream << endl;
		resultStream << concreteReadSt->second.readId << SEPARATOR << FWD_MAPPING << SEPARATOR
				<< chrName << SEPARATOR << relativePos2 << SEPARATOR;
		resultStream << concreteReadSt->second.readString << SEPARATOR
				<< concreteReadSt->second.readQuality << SEPARATOR;
		resultStream << N_TAG << forwardIter->distanceRight << SEPARATOR;
		if (forwardIter2 != forwardEnd) {
			resultStream << XA_Z_TAG;
		}
		for (; forwardIter2 != forwardEnd; forwardIter2++) {
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, forwardIter2->posRight, chrName, chrPos);
			chrPos++;
			resultStream << chrName << COMMA << FWD_MAPPING << chrPos << COMMA
					<< concreteReadSt->second.readString.size() << CIGAR << COMMA
					<< forwardIter2->distanceRight << SEMI_COLON;
		}
		for (; reverseIter2 != reverseEnd; reverseIter2++) {
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, reverseIter2->posRight, chrName, chrPos);
			chrPos++;
			resultStream << chrName << COMMA << REV_MAPPING << chrPos << COMMA
					<< concreteReadSt->second.reverseReadString.size() << CIGAR << COMMA
					<< reverseIter2->distanceRight << SEMI_COLON;
		}
	} else if (reverseIter != reverseEnd) {
		entered = true;
		// Write the first result.
		getChromosomeOffset(searcher->indexer->chromosomeNames,
				searcher->indexer->chromosomeLengths, reverseIter->posRight, chrName, relativePos);
		relativePos++;
		getChromosomeOffset(searcher->indexer->chromosomeNames,
				searcher->indexer->chromosomeLengths, reverseIter->pos, chrName, relativePos2);
		relativePos2++;
		resultStream << concreteReadSt->first.readId << SEPARATOR << REV_MAPPING << SEPARATOR
				<< chrName << SEPARATOR << relativePos << SEPARATOR;
		resultStream << concreteReadSt->second.reverseReadString << SEPARATOR
				<< concreteReadSt->second.reverseReadQuality << SEPARATOR;
		resultStream << N_TAG << reverseIter->distanceRight << SEPARATOR;
		reverseIter++;
		vector<PairedReadMapping>::iterator reverseIter2 = reverseIter;
		if (reverseIter != reverseEnd) {
			resultStream << XA_Z_TAG;
		}
		for (; reverseIter != reverseEnd; reverseIter++) {
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, reverseIter->posRight, chrName, chrPos);
			chrPos++;
			resultStream << chrName << COMMA << REV_MAPPING << chrPos << COMMA
					<< concreteReadSt->second.reverseReadString.size() << CIGAR << COMMA
					<< reverseIter->distanceRight << SEMI_COLON;
		}
		resultStream << endl;
		resultStream << concreteReadSt->second.readId << SEPARATOR << REV_MAPPING << SEPARATOR
				<< chrName << SEPARATOR << relativePos2 << SEPARATOR;
		resultStream << concreteReadSt->first.reverseReadString << SEPARATOR
				<< concreteReadSt->first.reverseReadQuality << SEPARATOR;
		resultStream << N_TAG << reverseIter->distance << SEPARATOR;
		if (reverseIter2 != reverseEnd) {
			resultStream << XA_Z_TAG;
		}
		for (; reverseIter2 != reverseEnd; reverseIter2++) {
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, reverseIter2->pos, chrName, chrPos);
			chrPos++;
			resultStream << chrName << COMMA << REV_MAPPING << chrPos << COMMA
					<< concreteReadSt->first.reverseReadString.size() << CIGAR << COMMA
					<< reverseIter2->distance << SEMI_COLON;
		}
	}
	if (entered) {
		resultStream << endl;
	}
}
void PairedEnhancedNativeFormatResultStringWriter::writeResultString(stringstream& resultStream, ReadInfo& readInfo, Searcher* searcher) {
	clog << "ERROR: PairedEnhancedNativeFormatResultStringWriter not implemented." << endl;
	exit(0);
}

void BowtieFormatResultStringWriter::writeResultString(stringstream& resultStream,
		ReadInfo& readInfo, Searcher* searcher) {
	SingleEndSearcher* concreteSearcher = static_cast<SingleEndSearcher*>(searcher);
	vector<ReadMapping>::iterator forwardIter = concreteSearcher->forwardResults.begin();
	vector<ReadMapping>::iterator forwardEnd = concreteSearcher->forwardResults.end();
	string chrName;
	unsigned relativePos = 0;
	char forFlag = '+';
	char revFlag = '-';
	string gencigar = boost::lexical_cast < string > (readInfo.readString.size());
	gencigar.append("M");
	if (conf.queryType == SELECT_HEURISTIC_INDEL) {
		if (!conf.genCigar) {
			gencigar = "*";
		}
	}
	if (concreteSearcher->forwardResults.size() == 0
			&& concreteSearcher->reverseResults.size() == 0) {

		resultStream << readInfo.readId << SEPARATOR << forFlag << SEPARATOR << UNAVAILABLE_1
				<< SEPARATOR << UNAVAILABLE_2 << SEPARATOR;
		resultStream << UNAVAILABLE_2 << SEPARATOR << UNAVAILABLE_1 << SEPARATOR << UNAVAILABLE_1
				<< SEPARATOR << UNAVAILABLE_2 << SEPARATOR;
		resultStream << UNAVAILABLE_2 << SEPARATOR << readInfo.readString << SEPARATOR
				<< readInfo.readQuality<<endl;

	} else {
		for (; forwardIter != forwardEnd; forwardIter++) {
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, forwardIter->pos, chrName,
					relativePos);
			relativePos++;
			resultStream << readInfo.readId << SEPARATOR << forFlag << SEPARATOR;
			resultStream << chrName << SEPARATOR << relativePos << SEPARATOR << mapQuality
					<< SEPARATOR;
			if (forwardIter->cigar != NULL) {
				resultStream << forwardIter->cigar->str() << SEPARATOR;
				delete forwardIter->cigar;
			} else {
				resultStream << gencigar << SEPARATOR;
			}
			resultStream << refNext << SEPARATOR << posNext << SEPARATOR;
			resultStream << templateLen << SEPARATOR << readInfo.readString << SEPARATOR
					<< readInfo.readQuality << SEPARATOR;
			resultStream << NM_TAG << forwardIter->distance << endl;
		}
		vector<ReadMapping>::iterator reverseIter = concreteSearcher->reverseResults.begin();
		vector<ReadMapping>::iterator reverseEnd = concreteSearcher->reverseResults.end();
		for (; reverseIter != reverseEnd; reverseIter++) {
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, reverseIter->pos, chrName,
					relativePos);
			relativePos++;
			samFlags += READ_REVERSE_MAPPED;
			resultStream << readInfo.readId << SEPARATOR << revFlag << SEPARATOR;
			resultStream << chrName << SEPARATOR << relativePos << SEPARATOR << mapQuality
					<< SEPARATOR;
			if (forwardIter->cigar != NULL) {
				resultStream << forwardIter->cigar->str() << SEPARATOR;
				delete forwardIter->cigar;
			} else {
				resultStream << gencigar << SEPARATOR;
			}
			resultStream << refNext << SEPARATOR << posNext << SEPARATOR;
			resultStream << templateLen << SEPARATOR << readInfo.reverseReadString << SEPARATOR
					<< readInfo.reverseReadQuality << SEPARATOR;
			resultStream << NM_TAG << reverseIter->distance << endl;
		}
	}
}

void PairedBowtieFormatResultStringWriter::writeResultString(stringstream& resultStream,
		ReadInfo& readInfo, Searcher* searcher) {
	PairedEndSearcher* concreteSearcher = static_cast<PairedEndSearcher*>(searcher);
	PairedEndReadInfo* concreteReadSt = static_cast<PairedEndReadInfo*>(&readInfo);
	vector<PairedReadMapping>::iterator forwardIter = concreteSearcher->forwardResults.begin();
	vector<PairedReadMapping>::iterator forwardEnd = concreteSearcher->forwardResults.end();
	vector<PairedReadMapping>::iterator reverseIter = concreteSearcher->reverseResults.begin();
	vector<PairedReadMapping>::iterator reverseEnd = concreteSearcher->reverseResults.end();
	string chrName;
	string chrName2;
	unsigned relativePos = 0;
	unsigned relativePos2 = 0;
	bool entered = false;
	char refNext = '=';
	string gencigar = boost::lexical_cast < string > (concreteReadSt->first.readString.size());
	gencigar.append("M");
	if (conf.queryType == SELECT_HEURISTIC_INDEL) {
		if (!conf.genCigar) {
			gencigar = "*";
		}
	}
	samFlags = 0;
	if (concreteSearcher->forwardResults.size() == 0
			&& concreteSearcher->reverseResults.size() == 0) {
		entered = true;
		samFlags = READ_PAIR_UNMAPPED1;
		resultStream << concreteReadSt->first.readId << SEPARATOR;
		resultStream << samFlags << SEPARATOR << UNAVAILABLE_1 << SEPARATOR << UNAVAILABLE_2
				<< SEPARATOR << UNAVAILABLE_2 << SEPARATOR << UNAVAILABLE_1 << SEPARATOR
				<< UNAVAILABLE_1 << SEPARATOR << UNAVAILABLE_2 << SEPARATOR << UNAVAILABLE_2
				<< SEPARATOR << concreteReadSt->first.readString << SEPARATOR
				<< concreteReadSt->first.readQuality << endl;
		samFlags = READ_PAIR_UNMAPPED2;
		resultStream << concreteReadSt->second.readId << SEPARATOR << samFlags << SEPARATOR
				<< UNAVAILABLE_1 << SEPARATOR << UNAVAILABLE_2 << SEPARATOR << UNAVAILABLE_2
				<< SEPARATOR << UNAVAILABLE_1 << SEPARATOR << UNAVAILABLE_1 << SEPARATOR
				<< UNAVAILABLE_2 << SEPARATOR << UNAVAILABLE_2 << SEPARATOR
				<< concreteReadSt->second.readString << SEPARATOR
				<< concreteReadSt->second.readQuality << endl;
	} else {
		while (forwardIter != forwardEnd) {
			// Write the first result.
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, forwardIter->pos, chrName,
					relativePos);
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, forwardIter->posRight, chrName2,
					relativePos2);
			entered = true;
			relativePos++;
			relativePos2++;
			int nextPos = relativePos2 - relativePos + concreteReadSt->first.readString.size();
			samFlags = READ_PAIR_FORWARD_MAPPED1;
			resultStream << concreteReadSt->first.readId << SEPARATOR;
			resultStream << samFlags << SEPARATOR << chrName << SEPARATOR << relativePos
					<< SEPARATOR;
			resultStream << mapQuality << SEPARATOR;
			if (forwardIter->cigar != NULL) {
				resultStream << forwardIter->cigar->str() << SEPARATOR;
				delete forwardIter->cigar;
			} else {
				resultStream << gencigar << SEPARATOR;
			}
			resultStream << refNext << SEPARATOR << relativePos2 << SEPARATOR;
			resultStream << nextPos << SEPARATOR << concreteReadSt->first.readString << SEPARATOR
					<< concreteReadSt->first.readQuality << SEPARATOR;
			resultStream << NM_TAG << forwardIter->distance << SEPARATOR;
			samFlags = READ_PAIR_FORWARD_MAPPED2;
			int nextPos2 = -nextPos;
			resultStream << concreteReadSt->second.readId << SEPARATOR << samFlags << SEPARATOR
					<< chrName << SEPARATOR << relativePos2 << SEPARATOR;
			resultStream << mapQuality << SEPARATOR;
			if (forwardIter->cigarRight != NULL) {
				resultStream << forwardIter->cigarRight->str() << SEPARATOR;
				delete forwardIter->cigarRight;
			} else {
				resultStream << gencigar << SEPARATOR;
			}
			resultStream << refNext << SEPARATOR << relativePos << SEPARATOR;
			resultStream << nextPos2 << SEPARATOR << concreteReadSt->second.readString << SEPARATOR
					<< concreteReadSt->second.readQuality << SEPARATOR;
			resultStream << NM_TAG << forwardIter->distanceRight << SEPARATOR;
			forwardIter++;
		}
		while (reverseIter != reverseEnd) {
			// Write the first result.
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, reverseIter->posRight, chrName,
					relativePos);
			relativePos++;
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, reverseIter->pos, chrName2,
					relativePos2);
			relativePos2++;
			entered = true;
			int nextPos = relativePos2 - relativePos - concreteReadSt->first.readString.size();
			samFlags = READ_PAIR_REVERSE_MAPPED1;
			resultStream << concreteReadSt->first.readId << SEPARATOR;
			resultStream << samFlags << SEPARATOR << chrName << SEPARATOR << relativePos
					<< SEPARATOR;
			resultStream << mapQuality << SEPARATOR;
			if (reverseIter->cigarRight != NULL) {
				resultStream << reverseIter->cigarRight->str() << SEPARATOR;
				delete reverseIter->cigarRight;
			} else {
				resultStream << gencigar << SEPARATOR;
			}
			resultStream << refNext << SEPARATOR << relativePos2 << SEPARATOR;
			resultStream << nextPos << SEPARATOR << concreteReadSt->second.reverseReadString
					<< SEPARATOR << concreteReadSt->second.reverseReadQuality << SEPARATOR;
			resultStream << NM_TAG << reverseIter->distanceRight << SEPARATOR;
			samFlags = READ_PAIR_REVERSE_MAPPED2;
			int nextPos2 = -nextPos;
			resultStream << concreteReadSt->second.readId << SEPARATOR << samFlags << SEPARATOR
					<< chrName << SEPARATOR << relativePos2 << SEPARATOR;
			resultStream << mapQuality << SEPARATOR;
			if (reverseIter->cigar != NULL) {
				resultStream << reverseIter->cigar->str() << SEPARATOR;
				delete reverseIter->cigar;
			} else {
				resultStream << gencigar << SEPARATOR;
			}
			resultStream << refNext << SEPARATOR << relativePos << SEPARATOR;
			resultStream << nextPos2 << SEPARATOR << concreteReadSt->first.reverseReadString
					<< SEPARATOR << concreteReadSt->first.reverseReadQuality << SEPARATOR;
			resultStream << NM_TAG << reverseIter->distance << SEPARATOR;
			reverseIter++;
		}
	}
}

void PairedEnhancedBowtieFormatResultStringWriter::writeResultString(stringstream& resultStream, ReadInfo& readInfo, Searcher* searcher) {
	clog << "ERROR: PairedEnhancedBowtieFormatResultStringWriter not implemented." << endl;
	exit(0);
}

void SAMFormatResultStringWriter::createInitialString(stringstream& resultStream,
		Indexer& indexer) const {
	if (conf.samOutputOptions == SAM_NO_HEADERS) {
		return;
	}
	// Writing HD TAG.
	resultStream << HD_TAG << SEPARATOR << VERSION_TAG << HOBBES_VERSION << SEPARATOR << SO_TAG
			<< "unsorted" << endl;
	if (conf.samOutputOptions == SAM_NO_SEQ) {
		resultStream << PROG_TAG << SEPARATOR << VERSION_TAG << HOBBES_VERSION << CMD_LINE_TAG
				<< "\"" << conf.commandLineInput << "\"" << endl;
		return;
	}
	// Writing SQ TAGS.
	unsigned currLen = 0;
	unsigned prevLen = 0;
	for (unsigned i = 0; i < indexer.chromosomeLengths.size(); i++) {
		currLen = indexer.chromosomeLengths[i];
		resultStream << SEQ_TAG << SEPARATOR << SEQNAME_TAG << indexer.chromosomeNames[i]
				<< SEPARATOR << SEQLEN_TAG << (currLen - prevLen) << endl;
		prevLen = currLen;
	}
	// Writing PROG TAG.
	resultStream << PROG_TAG << SEPARATOR << VERSION_TAG << HOBBES_VERSION << CMD_LINE_TAG << "\""
			<< conf.commandLineInput << "\"" << endl;
}

void SAMFormatResultStringWriter::writeResultString(stringstream& resultStream, ReadInfo& readInfo,
		Searcher* searcher) {
	SingleEndSearcher* concreteSearcher = static_cast<SingleEndSearcher*>(searcher);
	vector<ReadMapping>::iterator forwardIter = concreteSearcher->forwardResults.begin();
	vector<ReadMapping>::iterator forwardEnd = concreteSearcher->forwardResults.end();
	vector<ReadMapping>::iterator reverseIter = concreteSearcher->reverseResults.begin();
	vector<ReadMapping>::iterator reverseEnd = concreteSearcher->reverseResults.end();
	string chrName;
	unsigned relativePos = 0;
	bool entered = false;
	samFlags = 0;
	string gencigar = boost::lexical_cast < string > (readInfo.readString.size());
	gencigar.append("M");
	if (conf.queryType == SELECT_HEURISTIC_INDEL) {
		if (!conf.genCigar) {
			gencigar = "*";
		}
	}
	if (concreteSearcher->forwardResults.size() == 0
			&& concreteSearcher->reverseResults.size() == 0) {
		samFlags += READ_UNMAPPED;
		entered = true;
		
		resultStream << readInfo.readId << SEPARATOR << samFlags << SEPARATOR << UNAVAILABLE_1
				<< SEPARATOR << UNAVAILABLE_2 << SEPARATOR;
		resultStream << UNAVAILABLE_2 << SEPARATOR << UNAVAILABLE_1 << SEPARATOR << UNAVAILABLE_1
				<< SEPARATOR << UNAVAILABLE_2 << SEPARATOR;
		resultStream << UNAVAILABLE_2 << SEPARATOR << readInfo.readString << SEPARATOR
				<< readInfo.readQuality;
	} else {
		if (forwardIter != forwardEnd) {
			// Write the first result.
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, forwardIter->pos, chrName, relativePos);
			// For 1-based offset.
			relativePos++;
//			resultStream << readInfo.readString << SEPARATOR << readInfo.readQuality << endl;
			resultStream << readInfo.readId << SEPARATOR << samFlags << SEPARATOR << chrName
					<< SEPARATOR << relativePos << SEPARATOR;
			resultStream << mapQuality << SEPARATOR;
			if (forwardIter->cigar != NULL) {
				resultStream << forwardIter->cigar->str() << SEPARATOR;
				delete forwardIter->cigar;
			} else {
				resultStream << gencigar << SEPARATOR;
			}
			resultStream << refNext << SEPARATOR << posNext << SEPARATOR;
			resultStream << templateLen << SEPARATOR << readInfo.readString << SEPARATOR
					<< readInfo.readQuality << SEPARATOR;
			resultStream << NM_TAG << forwardIter->distance << SEPARATOR << XM_TAG
					<< forwardIter->distance << SEPARATOR;
			forwardIter++;
			if (forwardIter != forwardEnd) {
				resultStream << XA_Z_TAG;
			}
			for (; forwardIter != forwardEnd; forwardIter++) {
				entered = true;
				getChromosomeOffset(searcher->indexer->chromosomeNames,
						searcher->indexer->chromosomeLengths, forwardIter->pos, chrName,
						relativePos);
				relativePos++;
				resultStream << chrName << COMMA << FWD_MAPPING << relativePos << COMMA;
				if (forwardIter->cigar != NULL) {
					resultStream << forwardIter->cigar->str() << COMMA;
					delete forwardIter->cigar;
				} else {
					resultStream << gencigar << COMMA;
				}
				resultStream << forwardIter->distance << SEMI_COLON;
			}
			for (; reverseIter != reverseEnd; reverseIter++) {
				entered = true;
				getChromosomeOffset(searcher->indexer->chromosomeNames,
						searcher->indexer->chromosomeLengths, reverseIter->pos, chrName,
						relativePos);
				relativePos++;
				resultStream << chrName << COMMA << REV_MAPPING << relativePos << COMMA;
				if (reverseIter->cigar != NULL) {
					resultStream << reverseIter->cigar->str() << COMMA;
					delete reverseIter->cigar;
				} else {
					resultStream << gencigar << COMMA;
				}
				resultStream << reverseIter->distance << SEMI_COLON;
			}
		} else if (reverseIter != reverseEnd) {
			// Write the first result.
			samFlags += READ_REVERSE_MAPPED;
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, reverseIter->pos, chrName, relativePos);
			relativePos++;
//			resultStream << readInfo.readString << SEPARATOR << readInfo.readQuality << endl;
			resultStream << readInfo.readId << SEPARATOR << samFlags << SEPARATOR << chrName
					<< SEPARATOR << relativePos << SEPARATOR;
			resultStream << mapQuality << SEPARATOR;
			if (reverseIter->cigar != NULL) {
				resultStream << reverseIter->cigar->str() << SEPARATOR;
				delete reverseIter->cigar;
			} else {
				resultStream << gencigar << SEPARATOR;
			}
			resultStream << refNext << SEPARATOR << posNext << SEPARATOR;
			resultStream << templateLen << SEPARATOR << readInfo.reverseReadString << SEPARATOR
					<< readInfo.reverseReadQuality << SEPARATOR;
			resultStream << NM_TAG << reverseIter->distance << SEPARATOR << XM_TAG
					<< reverseIter->distance << SEPARATOR;
			reverseIter++;
			if (reverseIter != reverseEnd) {
				resultStream << XA_Z_TAG;
			}
			for (; reverseIter != reverseEnd; reverseIter++) {
				entered = true;
				getChromosomeOffset(searcher->indexer->chromosomeNames,
						searcher->indexer->chromosomeLengths, reverseIter->pos, chrName,
						relativePos);
				relativePos++;
				resultStream << chrName << COMMA << REV_MAPPING << relativePos << COMMA;
				if (reverseIter->cigar != NULL) {
					resultStream << reverseIter->cigar->str() << COMMA;
					delete reverseIter->cigar;
				} else {
					resultStream << gencigar << COMMA;
				}
				resultStream << reverseIter->distance << SEMI_COLON;
			}
		}
	}
	if (!entered) {
		resultStream << UNAVAILABLE_1 << endl;
	} else {
		resultStream << endl;
	}
}

void ResultStringWriter::getChromosomeOffset(vector<string>& chrNames, vector<unsigned>& chrLengths,
		unsigned absolutePos, string& chrName, unsigned& relativePos) const {
	chrName.clear();
	relativePos = 0;
	vector<unsigned>::iterator itr = upper_bound(chrLengths.begin(), chrLengths.end(), absolutePos);
	if (*itr > chrLengths[0]) {
		relativePos = (absolutePos - *(itr - 1));
	} else {
		relativePos = absolutePos;
	}
	chrName = chrNames[(itr - chrLengths.begin())];
}

void PairedSAMFormatResultStringWriter::createInitialString(stringstream& resultStream,
		Indexer& indexer) const {
	if (conf.samOutputOptions == SAM_NO_HEADERS) {
		return;
	}
	// Writing HD TAG.
	resultStream << HD_TAG << SEPARATOR << VERSION_TAG << HOBBES_VERSION << SEPARATOR << SO_TAG
			<< "unsorted" << endl;
	if (conf.samOutputOptions == SAM_NO_SEQ) {
		resultStream << PROG_TAG << SEPARATOR << VERSION_TAG << HOBBES_VERSION << CMD_LINE_TAG
				<< "\"" << conf.commandLineInput << "\"" << endl;
		return;
	}
	// Writing SQ TAGS.
	unsigned currLen = 0;
	unsigned prevLen = 0;
	for (unsigned i = 0; i < indexer.chromosomeLengths.size(); i++) {
		currLen = indexer.chromosomeLengths[i];
		resultStream << SEQ_TAG << SEPARATOR << SEQNAME_TAG << indexer.chromosomeNames[i]
				<< SEPARATOR << SEQLEN_TAG << (currLen - prevLen) << endl;
		prevLen = currLen;
	}
	// Writing PROG TAG.
	resultStream << PROG_TAG << SEPARATOR << VERSION_TAG << HOBBES_VERSION << CMD_LINE_TAG << "\""
			<< conf.commandLineInput << "\"" << endl;
}

void PairedEnhancedSAMFormatResultStringWriter::createInitialString(stringstream& resultStream,
		Indexer& indexer) const {
	if (conf.samOutputOptions == SAM_NO_HEADERS) {
		return;
	}
	// Writing HD TAG.
	resultStream << HD_TAG << SEPARATOR << VERSION_TAG << HOBBES_VERSION << SEPARATOR << SO_TAG
			<< "unsorted" << endl;
	if (conf.samOutputOptions == SAM_NO_SEQ) {
		resultStream << PROG_TAG << SEPARATOR << VERSION_TAG << HOBBES_VERSION << CMD_LINE_TAG
				<< "\"" << conf.commandLineInput << "\"" << endl;
		return;
	}
	// Writing SQ TAGS.
	unsigned currLen = 0;
	unsigned prevLen = 0;
	for (unsigned i = 0; i < indexer.chromosomeLengths.size(); i++) {
		currLen = indexer.chromosomeLengths[i];
		resultStream << SEQ_TAG << SEPARATOR << SEQNAME_TAG << indexer.chromosomeNames[i]
				<< SEPARATOR << SEQLEN_TAG << (currLen - prevLen) << endl;
		prevLen = currLen;
	}
	// Writing PROG TAG.
	resultStream << PROG_TAG << SEPARATOR << VERSION_TAG << HOBBES_VERSION << CMD_LINE_TAG << "\""
			<< conf.commandLineInput << "\"" << endl;
}

void PairedSAMFormatResultStringWriter::writeResultString(stringstream& resultStream,
		ReadInfo& readInfo, Searcher* searcher) {
	PairedEndSearcher* concreteSearcher = static_cast<PairedEndSearcher*>(searcher);
	PairedEndReadInfo* concreteReadSt = static_cast<PairedEndReadInfo*>(&readInfo);
	vector<PairedReadMapping>::iterator forwardIter = concreteSearcher->forwardResults.begin();
	vector<PairedReadMapping>::iterator forwardEnd = concreteSearcher->forwardResults.end();
	vector<PairedReadMapping>::iterator reverseIter = concreteSearcher->reverseResults.begin();
	vector<PairedReadMapping>::iterator reverseEnd = concreteSearcher->reverseResults.end();
	string chrName;
	string chrName2;
	unsigned relativePos = 0;
	unsigned relativePos2 = 0;
	unsigned chrPos = 0;
	bool entered = false;
	char refNext = '='; //unavailable
	string gencigar = boost::lexical_cast < string > (concreteReadSt->first.readString.size());
	gencigar.append("M");
	if (conf.queryType == SELECT_HEURISTIC_INDEL) {
		if (!conf.genCigar) {
			gencigar = "*";
		}
	}
	samFlags = 0;
	if (concreteSearcher->forwardResults.size() == 0
			&& concreteSearcher->reverseResults.size() == 0) {
		entered = true;
		samFlags = READ_PAIR_UNMAPPED1;
		resultStream << concreteReadSt->first.readId << SEPARATOR;
		resultStream << samFlags << SEPARATOR << UNAVAILABLE_1 << SEPARATOR << UNAVAILABLE_2
				<< SEPARATOR << UNAVAILABLE_2 << SEPARATOR << UNAVAILABLE_1 << SEPARATOR
				<< UNAVAILABLE_1 << SEPARATOR << UNAVAILABLE_2 << SEPARATOR << UNAVAILABLE_2
				<< SEPARATOR << concreteReadSt->first.readString << SEPARATOR
				<< concreteReadSt->first.readQuality << endl;
		samFlags = READ_PAIR_UNMAPPED2;
		resultStream << concreteReadSt->second.readId << SEPARATOR << samFlags << SEPARATOR
				<< UNAVAILABLE_1 << SEPARATOR << UNAVAILABLE_2 << SEPARATOR << UNAVAILABLE_2
				<< SEPARATOR << UNAVAILABLE_1 << SEPARATOR << UNAVAILABLE_1 << SEPARATOR
				<< UNAVAILABLE_2 << SEPARATOR << UNAVAILABLE_2 << SEPARATOR
				<< concreteReadSt->second.readString << SEPARATOR
				<< concreteReadSt->second.readQuality << endl;
	} else {
		if (forwardIter != forwardEnd) {
			// Write the first result.
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, forwardIter->pos, chrName, relativePos);
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, forwardIter->posRight, chrName2,
					relativePos2);
			entered = true;
			// For 1-based offset.
			relativePos++;
			relativePos2++;
			int nextPos = relativePos2 - relativePos + concreteReadSt->first.readString.size();
			samFlags = READ_PAIR_FORWARD_MAPPED1;
			resultStream << concreteReadSt->first.readId << SEPARATOR;
			resultStream << samFlags << SEPARATOR << chrName << SEPARATOR << relativePos
					<< SEPARATOR;
			resultStream << mapQuality << SEPARATOR;
			if (forwardIter->cigar != NULL) {
				resultStream << forwardIter->cigar->str() << SEPARATOR;
				delete forwardIter->cigar;
			} else {
				resultStream << gencigar << SEPARATOR;
			}
			resultStream << refNext << SEPARATOR << relativePos2 << SEPARATOR;
			resultStream << nextPos << SEPARATOR << concreteReadSt->first.readString << SEPARATOR
					<< concreteReadSt->first.readQuality << SEPARATOR;
			resultStream << NM_TAG << forwardIter->distance;// << SEPARATOR; //XA:i:0	MD:Z:50	NM:i:0
			vector<PairedReadMapping>::iterator forwardIter2 = forwardIter;
			vector<PairedReadMapping>::iterator reverseIter2 = reverseIter;
			forwardIter++;
			if (forwardIter != forwardEnd) {
				//resultStream << XA_Z_TAG;
				resultStream << SEPARATOR << XA_Z_TAG;
			}
			for (; forwardIter != forwardEnd; forwardIter++) {
				getChromosomeOffset(searcher->indexer->chromosomeNames,
						searcher->indexer->chromosomeLengths, forwardIter->pos, chrName, chrPos);
				chrPos++;
				resultStream << chrName << COMMA << FWD_MAPPING << chrPos << COMMA;
				if (forwardIter->cigar != NULL) {
					resultStream << forwardIter->cigar->str() << COMMA;
					delete forwardIter->cigar;
				} else {
					resultStream << gencigar << COMMA;
				}
				resultStream << forwardIter->distance << SEMI_COLON;
			}
			for (; reverseIter != reverseEnd; reverseIter++) {
				getChromosomeOffset(searcher->indexer->chromosomeNames,
						searcher->indexer->chromosomeLengths, reverseIter->posRight, chrName, chrPos);
				chrPos++;
				resultStream << chrName << COMMA << REV_MAPPING << chrPos << COMMA;
				if (reverseIter->cigarRight != NULL) {
					resultStream << reverseIter->cigarRight->str() << COMMA;
					delete reverseIter->cigarRight;
				} else {
					resultStream << gencigar << COMMA;
				}
				resultStream << reverseIter->distanceRight << SEMI_COLON;
			}
			resultStream << endl;
			samFlags = READ_PAIR_FORWARD_MAPPED2;
			int nextPos2 = -nextPos;
			resultStream << concreteReadSt->second.readId << SEPARATOR << samFlags << SEPARATOR
					<< chrName2 << SEPARATOR << relativePos2 << SEPARATOR;
			resultStream << mapQuality << SEPARATOR;
			if (forwardIter2->cigarRight != NULL) {
				resultStream << forwardIter2->cigarRight->str() << SEPARATOR;
				delete forwardIter2->cigarRight;
			} else {
				resultStream << gencigar << SEPARATOR;
			}
			resultStream << refNext << SEPARATOR << relativePos << SEPARATOR;
			resultStream << nextPos2 << SEPARATOR << concreteReadSt->second.readString << SEPARATOR
					<< concreteReadSt->second.readQuality << SEPARATOR;
			resultStream << NM_TAG << forwardIter2->distanceRight;// << SEPARATOR;
			forwardIter2++;
			if (forwardIter2 != forwardEnd) {
				resultStream << SEPARATOR << XA_Z_TAG;
			}
			for (; forwardIter2 != forwardEnd; forwardIter2++) {
				getChromosomeOffset(searcher->indexer->chromosomeNames,
						searcher->indexer->chromosomeLengths, forwardIter2->posRight, chrName, chrPos);
				chrPos++;
				resultStream << chrName << COMMA << FWD_MAPPING << chrPos << COMMA;
				if (forwardIter2->cigarRight != NULL) {
					resultStream << forwardIter2->cigarRight->str() << COMMA;
					delete forwardIter2->cigarRight;
				} else {
					resultStream << gencigar << COMMA;
				}
				resultStream << forwardIter2->distanceRight << SEMI_COLON;
			}
			for (; reverseIter2 != reverseEnd; reverseIter2++) {
				getChromosomeOffset(searcher->indexer->chromosomeNames,
						searcher->indexer->chromosomeLengths, reverseIter2->pos, chrName, chrPos);
				chrPos++;
				resultStream << chrName << COMMA << REV_MAPPING << chrPos << COMMA;
				if (reverseIter2->cigar != NULL) {
					resultStream << reverseIter2->cigar->str() << COMMA;
					delete reverseIter2->cigar;
				} else {
					resultStream << gencigar << COMMA;
				}
				resultStream << reverseIter2->distance << SEMI_COLON;
			}
			resultStream << endl;
		} else if (reverseIter != reverseEnd) {
			// Write the first result.
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, reverseIter->posRight, chrName, relativePos);
			relativePos++;
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, reverseIter->pos, chrName2, relativePos2);
			relativePos2++;
			entered = true;
			int nextPos = relativePos2 - relativePos - concreteReadSt->first.readString.size();
			samFlags = READ_PAIR_REVERSE_MAPPED1;
			resultStream << concreteReadSt->first.readId << SEPARATOR;
			resultStream << samFlags << SEPARATOR << chrName << SEPARATOR << relativePos
					<< SEPARATOR;
			resultStream << mapQuality << SEPARATOR;
			if (reverseIter->cigarRight != NULL) {
				resultStream << reverseIter->cigarRight->str() << SEPARATOR;
				delete reverseIter->cigarRight;
			} else {
				resultStream << gencigar << SEPARATOR;
			}
			resultStream << refNext << SEPARATOR << relativePos2 << SEPARATOR;
			resultStream << nextPos << SEPARATOR << concreteReadSt->second.reverseReadString
					<< SEPARATOR << concreteReadSt->second.reverseReadQuality << SEPARATOR;
			resultStream << NM_TAG << reverseIter->distanceRight;// << SEPARATOR;
			vector<PairedReadMapping>::iterator reverseIter2 = reverseIter;
			reverseIter++;
			if (reverseIter != reverseEnd) {
				resultStream << SEPARATOR << XA_Z_TAG;
			}
			for (; reverseIter != reverseEnd; reverseIter++) {
				getChromosomeOffset(searcher->indexer->chromosomeNames,
						searcher->indexer->chromosomeLengths, reverseIter->posRight, chrName, chrPos);
				chrPos++;
				resultStream << chrName << COMMA << REV_MAPPING << chrPos << COMMA;
				if (reverseIter->cigarRight != NULL) {
					resultStream << reverseIter->cigarRight->str() << COMMA;
					delete reverseIter->cigarRight;
				} else {
					resultStream << gencigar << COMMA;
				}
				resultStream << reverseIter->distanceRight << SEMI_COLON;
			}
			resultStream << endl;
			samFlags = READ_PAIR_REVERSE_MAPPED2;
			int nextPos2 = -nextPos;
			resultStream << concreteReadSt->second.readId << SEPARATOR << samFlags << SEPARATOR
					<< chrName2 << SEPARATOR << relativePos2 << SEPARATOR;
			resultStream << mapQuality << SEPARATOR;
			if (reverseIter2->cigar) {
				resultStream << reverseIter2->cigar->str() << SEPARATOR;
				delete reverseIter2->cigar;
			} else {
				resultStream << gencigar << SEPARATOR;
			}
			resultStream << refNext << SEPARATOR << relativePos << SEPARATOR;
			resultStream << nextPos2 << SEPARATOR << concreteReadSt->first.reverseReadString
					<< SEPARATOR << concreteReadSt->first.reverseReadQuality << SEPARATOR;
			resultStream << NM_TAG << reverseIter2->distance;// << SEPARATOR;
			reverseIter2++;
			if (reverseIter2 != reverseEnd) {
				resultStream << SEPARATOR << XA_Z_TAG;
			}
			for (; reverseIter2 != reverseEnd; reverseIter2++) {
				getChromosomeOffset(searcher->indexer->chromosomeNames,
						searcher->indexer->chromosomeLengths, reverseIter2->pos, chrName, chrPos);
				chrPos++;
				resultStream << chrName << COMMA << REV_MAPPING << chrPos << COMMA;
				if (reverseIter2->cigar != NULL) {
					resultStream << reverseIter2->cigar->str() << COMMA;
					delete reverseIter2->cigar;
				} else {
					resultStream << gencigar << COMMA;
				}
				resultStream << reverseIter2->distance << SEMI_COLON;
			}
			resultStream << endl;
		}
	}
}

void PairedEnhancedSAMFormatResultStringWriter::writeResultString(stringstream& resultStream,
		ReadInfo& readInfo, Searcher* searcher) {
	PairedEndSearcherEnhanced* concreteSearcher = static_cast<PairedEndSearcherEnhanced*>(searcher);
	PairedEndReadInfo* concreteReadSt = static_cast<PairedEndReadInfo*>(&readInfo);
	vector<PairedReadMapping>::iterator forwardIter = concreteSearcher->forwardResults.begin();
	vector<PairedReadMapping>::iterator forwardEnd = concreteSearcher->forwardResults.end();
	vector<PairedReadMapping>::iterator reverseIter = concreteSearcher->reverseResults.begin();
	vector<PairedReadMapping>::iterator reverseEnd = concreteSearcher->reverseResults.end();
	string chrName;
	unsigned relativePos = 0;
	unsigned relativePos2 = 0;
	unsigned chrPos = 0;
	bool entered = false;
	char refNext = '=';
	samFlags = 0;
	string gencigar = boost::lexical_cast < string > (concreteReadSt->first.readString.size());
	gencigar.append("M");
	if (conf.queryType == SELECT_HEURISTIC_INDEL) {
		if (!conf.genCigar) {
			gencigar = "*";
		}
	}
	if (concreteSearcher->forwardResults.size() != 0
			|| concreteSearcher->reverseResults.size() != 0) {
		if (forwardIter != forwardEnd) {
			// Write the first result/
			entered = true;
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, forwardIter->pos, chrName, relativePos);
			// For 1-based offset.
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, forwardIter->posRight, chrName, relativePos2);
			relativePos++;
			relativePos2++;
			int nextPos = relativePos2 - relativePos + concreteReadSt->first.readString.size();
			samFlags = READ_PAIR_FORWARD_MAPPED1;
			resultStream << concreteReadSt->first.readId << SEPARATOR;
			resultStream << samFlags << SEPARATOR << chrName << SEPARATOR << relativePos
					<< SEPARATOR;
			resultStream << mapQuality << SEPARATOR;
			if (forwardIter->cigar != NULL) {
				resultStream << forwardIter->cigar->str() << SEPARATOR;
			} else {
				resultStream << gencigar << SEPARATOR;
			}
			resultStream << refNext << SEPARATOR << relativePos2 << SEPARATOR;
			resultStream << nextPos << SEPARATOR << concreteReadSt->first.readString << SEPARATOR
					<< concreteReadSt->first.readQuality << SEPARATOR;
			resultStream << NM_TAG << forwardIter->distance << SEPARATOR; //XA:i:0	MD:Z:50	NM:i:0
			vector<PairedReadMapping>::iterator forwardIter2 = forwardIter;
			vector<PairedReadMapping>::iterator reverseIter2 = reverseIter;
			forwardIter++;
			if (forwardIter != forwardEnd) {
				resultStream << SEPARATOR << XA_Z_TAG;
			}
			for (; forwardIter != forwardEnd; forwardIter++) {
				getChromosomeOffset(searcher->indexer->chromosomeNames,
						searcher->indexer->chromosomeLengths, forwardIter->pos, chrName, chrPos);
				chrPos++;
				resultStream << chrName << COMMA << FWD_MAPPING << chrPos << COMMA;
				if (forwardIter->cigar != NULL) {
					resultStream << forwardIter->cigar->str() << COMMA;
				} else {
					resultStream << gencigar << COMMA;
				}
				resultStream << forwardIter->distance << SEMI_COLON;
			}
			for (; reverseIter != reverseEnd; reverseIter++) {
				getChromosomeOffset(searcher->indexer->chromosomeNames,
						searcher->indexer->chromosomeLengths, reverseIter->posRight, chrName, chrPos);
				chrPos++;
				resultStream << chrName << COMMA << REV_MAPPING << chrPos << COMMA;
				if (reverseIter->cigarRight != NULL) {
					resultStream << reverseIter->cigarRight->str() << COMMA;
				} else {
					resultStream << gencigar << COMMA;
				}
				resultStream << reverseIter->distanceRight << SEMI_COLON;
			}
			resultStream << endl;
			samFlags = READ_PAIR_FORWARD_MAPPED2;
			int nextPos2 = -nextPos;
			resultStream << concreteReadSt->second.readId << SEPARATOR << samFlags << SEPARATOR
					<< chrName << SEPARATOR << relativePos2 << SEPARATOR;
			resultStream << mapQuality << SEPARATOR;
			if (forwardIter2->cigarRight != NULL) {
				resultStream << forwardIter2->cigarRight->str() << SEPARATOR;
			} else {
				resultStream << gencigar << SEPARATOR;
			}
			resultStream << refNext << SEPARATOR << relativePos << SEPARATOR;
			resultStream << nextPos2 << SEPARATOR << concreteReadSt->second.readString << SEPARATOR
					<< concreteReadSt->second.readQuality << SEPARATOR;
			resultStream << NM_TAG << forwardIter2->distanceRight << SEPARATOR;
			if (forwardIter2 != forwardEnd) {
				resultStream << SEPARATOR << XA_Z_TAG;
			}
			for (; forwardIter2 != forwardEnd; forwardIter2++) {
				getChromosomeOffset(searcher->indexer->chromosomeNames,
						searcher->indexer->chromosomeLengths, forwardIter2->posRight, chrName, chrPos);
				chrPos++;
				resultStream << chrName << COMMA << FWD_MAPPING << chrPos << COMMA;
				if (forwardIter2->cigarRight != NULL) {
					resultStream << forwardIter2->cigarRight->str() << COMMA;
				} else {
					resultStream << gencigar << COMMA;
				}
				resultStream << forwardIter2->distanceRight << SEMI_COLON;
			}
			for (; reverseIter2 != reverseEnd; reverseIter2++) {
				getChromosomeOffset(searcher->indexer->chromosomeNames,
						searcher->indexer->chromosomeLengths, reverseIter2->posRight, chrName, chrPos);
				chrPos++;
				resultStream << chrName << COMMA << REV_MAPPING << chrPos << COMMA;
				if (reverseIter2->cigar != NULL) {
					resultStream << reverseIter2->cigar->str() << COMMA;
				} else {
					resultStream << gencigar << COMMA;
				}
				resultStream << reverseIter2->distance << SEMI_COLON;
			}
			resultStream << endl;
		} else if (reverseIter != reverseEnd) {
			entered = true;
			// Write the first result.
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, reverseIter->posRight, chrName, relativePos);
			relativePos++;
			getChromosomeOffset(searcher->indexer->chromosomeNames,
					searcher->indexer->chromosomeLengths, reverseIter->pos, chrName, relativePos2);
			relativePos2++;
			int nextPos = relativePos2 - relativePos - concreteReadSt->first.readString.size();
			samFlags = READ_PAIR_REVERSE_MAPPED1;
			resultStream << concreteReadSt->first.readId << SEPARATOR;
			resultStream << samFlags << SEPARATOR << chrName << SEPARATOR << relativePos
					<< SEPARATOR;
			resultStream << mapQuality << SEPARATOR;
			if (reverseIter->cigarRight != NULL) {
				resultStream << reverseIter->cigarRight->str() << SEPARATOR;
			} else {
				resultStream << gencigar << SEPARATOR;
			}
			resultStream << refNext << SEPARATOR << relativePos2 << SEPARATOR;
			resultStream << nextPos << SEPARATOR << concreteReadSt->second.reverseReadString
					<< SEPARATOR << concreteReadSt->second.reverseReadQuality << SEPARATOR;
			resultStream << NM_TAG << reverseIter->distanceRight << SEPARATOR;
			vector<PairedReadMapping>::iterator reverseIter2 = reverseIter;
			reverseIter++;
			if (reverseIter != reverseEnd) {
				resultStream << SEPARATOR << XA_Z_TAG;
			}
			for (; reverseIter != reverseEnd; reverseIter++) {
				getChromosomeOffset(searcher->indexer->chromosomeNames,
						searcher->indexer->chromosomeLengths, reverseIter->posRight, chrName, chrPos);
				chrPos++;
				resultStream << chrName << COMMA << REV_MAPPING << chrPos << COMMA;
				if (reverseIter->cigarRight != NULL) {
					resultStream << reverseIter->cigarRight->str() << COMMA;
				} else {
					resultStream << gencigar << COMMA;
				}
				resultStream << reverseIter->distanceRight << SEMI_COLON;
			}
			resultStream << endl;
			samFlags = READ_PAIR_REVERSE_MAPPED2;
			int nextPos2 = -nextPos;
			resultStream << concreteReadSt->second.readId << SEPARATOR << samFlags << SEPARATOR
					<< chrName << SEPARATOR << relativePos2 << SEPARATOR;
			resultStream << mapQuality << SEPARATOR;
			if (reverseIter2->cigar) {
				resultStream << reverseIter2->cigar->str() << SEPARATOR;
			} else {
				resultStream << gencigar << SEPARATOR;
			}
			resultStream << refNext << SEPARATOR << relativePos << SEPARATOR;
			resultStream << nextPos2 << SEPARATOR << concreteReadSt->first.reverseReadString
					<< SEPARATOR << concreteReadSt->first.reverseReadQuality << SEPARATOR;
			resultStream << NM_TAG << reverseIter2->distance << SEPARATOR;
			if (reverseIter2 != reverseEnd) {
				resultStream << SEPARATOR << XA_Z_TAG;
			}
			for (; reverseIter2 != reverseEnd; reverseIter2++) {
				getChromosomeOffset(searcher->indexer->chromosomeNames,
						searcher->indexer->chromosomeLengths, reverseIter2->pos, chrName, chrPos);
				chrPos++;
				resultStream << chrName << COMMA << REV_MAPPING << chrPos << COMMA;
				if (reverseIter2->cigar != NULL) {
					resultStream << reverseIter2->cigar->str() << COMMA;
				} else {
					resultStream << gencigar << COMMA;
				}
				resultStream << reverseIter2->distance << SEMI_COLON;
			}
			resultStream << endl;
		}
	} else {
		// Go to individual mapping result output.
		mapQuality = 0;
		int nextPos = 0;
		if (concreteSearcher->individualForwardResults1.size() != 0
				|| concreteSearcher->individualReverseResults2.size() != 0) { // read1 has at least a hit
			vector<ReadMapping>::iterator forwardIter =
					concreteSearcher->individualForwardResults1.begin();
			vector<ReadMapping>::iterator forwardEnd =
					concreteSearcher->individualForwardResults1.end();
			vector<ReadMapping>::iterator reverseIter =
					concreteSearcher->individualReverseResults2.begin();
			vector<ReadMapping>::iterator reverseEnd =
					concreteSearcher->individualReverseResults2.end();
			if (forwardIter != forwardEnd) {
				// Write the first result.
				getChromosomeOffset(searcher->indexer->chromosomeNames,
						searcher->indexer->chromosomeLengths, forwardIter->pos, chrName,
						relativePos);
				// For 1-based offset.
				relativePos++;
				samFlags = READ_PAIR_FIRST_FORWARD_MAPPED;
				resultStream << concreteReadSt->first.readId << SEPARATOR << samFlags << SEPARATOR
						<< chrName << SEPARATOR << relativePos << SEPARATOR;
				resultStream << mapQuality << SEPARATOR;
				if (reverseIter->cigar != NULL) {
					resultStream << reverseIter->cigar->str() << SEPARATOR;
				} else {
					resultStream << gencigar << SEPARATOR;
				}
				resultStream << refNext << SEPARATOR << templateLen << SEPARATOR;
				resultStream << nextPos << SEPARATOR << concreteReadSt->first.readString
						<< SEPARATOR << concreteReadSt->first.readQuality << SEPARATOR;
				resultStream << NM_TAG << forwardIter->distance << SEPARATOR;
				forwardIter++;
				if (forwardIter != forwardEnd) {
					resultStream << SEPARATOR << XA_Z_TAG;
				}
				for (; forwardIter != forwardEnd; forwardIter++) {
					getChromosomeOffset(searcher->indexer->chromosomeNames,
							searcher->indexer->chromosomeLengths, forwardIter->pos, chrName,
							relativePos);
					relativePos++;
					resultStream << chrName << COMMA << REV_MAPPING << relativePos << COMMA;
					if (forwardIter->cigar != NULL) {
						resultStream << forwardIter->cigar->str() << COMMA;
					} else {
						resultStream << gencigar << COMMA;
					}
					resultStream << forwardIter->distance << SEMI_COLON;
				}
				for (; reverseIter != reverseEnd; reverseIter++) {
					getChromosomeOffset(searcher->indexer->chromosomeNames,
							searcher->indexer->chromosomeLengths, reverseIter->pos, chrName,
							relativePos);
					relativePos++;
					resultStream << chrName << COMMA << REV_MAPPING << relativePos << COMMA;
					if (reverseIter->cigar != NULL) {
						resultStream << reverseIter->cigar->str() << COMMA;
					} else {
						resultStream << gencigar << COMMA;
					}
					resultStream << reverseIter->distance << SEMI_COLON;
				}
				resultStream << endl;
			} else if (reverseIter != reverseEnd) {
				// Write the first result.
				samFlags = READ_PAIR_FIRST_REVERSE_MAPPED;
				getChromosomeOffset(searcher->indexer->chromosomeNames,
						searcher->indexer->chromosomeLengths, reverseIter->pos, chrName,
						relativePos);
				relativePos++;
				resultStream << concreteReadSt->first.readId << SEPARATOR << samFlags << SEPARATOR
						<< chrName << SEPARATOR << relativePos << SEPARATOR;
				resultStream << mapQuality << SEPARATOR;
				if (reverseIter->cigar != NULL) {
					resultStream << reverseIter->cigar->str() << SEPARATOR;
				} else {
					resultStream << gencigar << SEPARATOR;
				}
				resultStream << refNext << SEPARATOR << templateLen << SEPARATOR;
				resultStream << nextPos << SEPARATOR << concreteReadSt->second.reverseReadString
						<< SEPARATOR << concreteReadSt->second.reverseReadQuality << SEPARATOR;
				resultStream << NM_TAG << reverseIter->distance << SEPARATOR;
				reverseIter++;
				if (reverseIter != reverseEnd) {
					resultStream << SEPARATOR << XA_Z_TAG;
				}
				for (; reverseIter != reverseEnd; reverseIter++) {
					getChromosomeOffset(searcher->indexer->chromosomeNames,
							searcher->indexer->chromosomeLengths, reverseIter->pos, chrName,
							relativePos);
					relativePos++;
					resultStream << chrName << COMMA << REV_MAPPING << relativePos << COMMA;
					if (reverseIter->cigar != NULL) {
						resultStream << reverseIter->cigar->str() << COMMA;
					} else {
						resultStream << gencigar << COMMA;
					}
					resultStream << reverseIter->distance << SEMI_COLON;
				}
				resultStream << endl;
			}
			if (concreteSearcher->individualForwardResults2.size() != 0
					|| concreteSearcher->individualReverseResults1.size() != 0) { // read1 has at least a hit
				vector<ReadMapping>::iterator forwardIter =
						concreteSearcher->individualForwardResults2.begin();
				vector<ReadMapping>::iterator forwardEnd =
						concreteSearcher->individualForwardResults2.end();
				vector<ReadMapping>::iterator reverseIter =
						concreteSearcher->individualReverseResults1.begin();
				vector<ReadMapping>::iterator reverseEnd =
						concreteSearcher->individualReverseResults1.end();
				if (forwardIter != forwardEnd) {
					// Write the first result.
					getChromosomeOffset(searcher->indexer->chromosomeNames,
							searcher->indexer->chromosomeLengths, forwardIter->pos, chrName,
							relativePos);
					// For 1-based offset.
					relativePos++;
					samFlags = READ_PAIR_SECOND_FORWARD_MAPPED;
					resultStream << concreteReadSt->second.readId << SEPARATOR << samFlags
							<< SEPARATOR << chrName << SEPARATOR << relativePos << SEPARATOR;
					resultStream << mapQuality << SEPARATOR;
					if (forwardIter->cigar != NULL) {
						resultStream << forwardIter->cigar->str() << SEPARATOR;
					} else {
						resultStream << gencigar << SEPARATOR;
					}
					resultStream << refNext << SEPARATOR << templateLen << SEPARATOR;
					resultStream << nextPos << SEPARATOR << concreteReadSt->second.readString
							<< SEPARATOR << concreteReadSt->second.readQuality << SEPARATOR;
					resultStream << NM_TAG << forwardIter->distance << SEPARATOR;
					forwardIter++;
					if (forwardIter != forwardEnd) {
						resultStream << SEPARATOR << XA_Z_TAG;
					}
					for (; forwardIter != forwardEnd; forwardIter++) {
						getChromosomeOffset(searcher->indexer->chromosomeNames,
								searcher->indexer->chromosomeLengths, forwardIter->pos, chrName,
								relativePos);
						relativePos++;
						resultStream << chrName << COMMA << REV_MAPPING << relativePos << COMMA;
						if (forwardIter->cigar != NULL) {
							resultStream << forwardIter->cigar->str() << COMMA;
						} else {
							resultStream << gencigar << COMMA;
						}
						resultStream << forwardIter->distance << SEMI_COLON;
					}
					for (; reverseIter != reverseEnd; reverseIter++) {
						getChromosomeOffset(searcher->indexer->chromosomeNames,
								searcher->indexer->chromosomeLengths, reverseIter->pos, chrName,
								relativePos);
						relativePos++;
						resultStream << chrName << COMMA << REV_MAPPING << relativePos << COMMA;
						if (reverseIter->cigar != NULL) {
							resultStream << reverseIter->cigar->str() << COMMA;
						} else {
							resultStream << gencigar << COMMA;
						}
						resultStream << reverseIter->distance << SEMI_COLON;
					}
					resultStream << endl;
				} else if (reverseIter != reverseEnd) {
					// Write the first result.
					samFlags = READ_PAIR_SECOND_REVERSE_MAPPED;
					getChromosomeOffset(searcher->indexer->chromosomeNames,
							searcher->indexer->chromosomeLengths, reverseIter->pos, chrName,
							relativePos);
					relativePos++;
					resultStream << concreteReadSt->second.readId << SEPARATOR << samFlags
							<< SEPARATOR << chrName << SEPARATOR << relativePos << SEPARATOR;
					resultStream << mapQuality << SEPARATOR;
					if (reverseIter->cigar != NULL) {
						resultStream << reverseIter->cigar->str() << SEPARATOR;
					} else {
						resultStream << gencigar << SEPARATOR;
					}
					resultStream << refNext << SEPARATOR;
					resultStream << templateLen << SEPARATOR
							<< concreteReadSt->first.reverseReadString << SEPARATOR
							<< concreteReadSt->first.reverseReadQuality << SEPARATOR;
					resultStream << NM_TAG << reverseIter->distance << SEPARATOR;
					reverseIter++;
					if (reverseIter != reverseEnd) {
						resultStream << SEPARATOR << XA_Z_TAG;
					}
					for (; reverseIter != reverseEnd; reverseIter++) {
						getChromosomeOffset(searcher->indexer->chromosomeNames,
								searcher->indexer->chromosomeLengths, reverseIter->pos, chrName,
								relativePos);
						relativePos++;
						resultStream << chrName << COMMA << REV_MAPPING << relativePos << COMMA;
						if (reverseIter->cigar != NULL) {
							resultStream << reverseIter->cigar->str() << COMMA;
						} else {
							resultStream << gencigar << COMMA;
						}
						resultStream << reverseIter->distance << SEMI_COLON;
					}
					resultStream << endl;
				}
			} else {
				// Second read no mapping.
				samFlags = READ_PAIR_SECOND_UNMAPPED;
				resultStream << concreteReadSt->second.readId << SEPARATOR << samFlags << SEPARATOR
						<< UNAVAILABLE_1 << SEPARATOR << UNAVAILABLE_2 << SEPARATOR << UNAVAILABLE_2
						<< SEPARATOR << UNAVAILABLE_1 << SEPARATOR << UNAVAILABLE_1 << SEPARATOR
						<< UNAVAILABLE_2 << SEPARATOR << UNAVAILABLE_2 << SEPARATOR
						<< concreteReadSt->second.readString << SEPARATOR
						<< concreteReadSt->second.readQuality << endl;
			}
		} else {
			// First read no mapping.
			samFlags = READ_PAIR_FIRST_UNMAPPED;
			resultStream << concreteReadSt->first.readId << SEPARATOR << samFlags << SEPARATOR
					<< UNAVAILABLE_1 << SEPARATOR << UNAVAILABLE_2 << SEPARATOR << UNAVAILABLE_2
					<< SEPARATOR << UNAVAILABLE_1 << SEPARATOR << UNAVAILABLE_1 << SEPARATOR
					<< UNAVAILABLE_2 << SEPARATOR << UNAVAILABLE_2 << SEPARATOR
					<< concreteReadSt->first.readString << SEPARATOR
					<< concreteReadSt->first.readQuality << endl;
			if (concreteSearcher->individualForwardResults2.size() != 0
					|| concreteSearcher->individualReverseResults1.size() != 0) { // read1 has at least a hit
				vector<ReadMapping>::iterator forwardIter =
						concreteSearcher->individualForwardResults2.begin();
				vector<ReadMapping>::iterator forwardEnd =
						concreteSearcher->individualForwardResults2.end();
				vector<ReadMapping>::iterator reverseIter =
						concreteSearcher->individualReverseResults1.begin();
				vector<ReadMapping>::iterator reverseEnd =
						concreteSearcher->individualReverseResults1.end();
				if (forwardIter != forwardEnd) {
					// Write the first result
					getChromosomeOffset(searcher->indexer->chromosomeNames,
							searcher->indexer->chromosomeLengths, forwardIter->pos, chrName,
							relativePos);
					// For 1-based offset
					relativePos++;
					samFlags = READ_PAIR_SECOND_FORWARD_MAPPED;
					resultStream << concreteReadSt->second.readId << SEPARATOR << samFlags
							<< SEPARATOR << chrName << SEPARATOR << relativePos << SEPARATOR;
					resultStream << mapQuality << SEPARATOR;
					if (forwardIter->cigar != NULL) {
						resultStream << forwardIter->cigar->str() << SEPARATOR;
					} else {
						resultStream << gencigar << SEPARATOR;
					}
					resultStream << refNext << SEPARATOR;
					resultStream << templateLen << SEPARATOR << concreteReadSt->second.readString
							<< SEPARATOR << concreteReadSt->second.readQuality << SEPARATOR;
					resultStream << NM_TAG << forwardIter->distance << SEPARATOR;
					forwardIter++;
					if (forwardIter != forwardEnd) {
						resultStream << SEPARATOR << XA_Z_TAG;
					}
					for (; forwardIter != forwardEnd; forwardIter++) {
						getChromosomeOffset(searcher->indexer->chromosomeNames,
								searcher->indexer->chromosomeLengths, forwardIter->pos, chrName,
								relativePos);
						relativePos++;
						resultStream << chrName << COMMA << REV_MAPPING << relativePos << COMMA;
						if (forwardIter->cigar != NULL) {
							resultStream << forwardIter->cigar->str() << COMMA;
						} else {
							resultStream << gencigar << SEPARATOR;
						}
						resultStream << forwardIter->distance << SEMI_COLON;
					}
					for (; reverseIter != reverseEnd; reverseIter++) {
						getChromosomeOffset(searcher->indexer->chromosomeNames,
								searcher->indexer->chromosomeLengths, reverseIter->pos, chrName,
								relativePos);
						relativePos++;
						resultStream << chrName << COMMA << REV_MAPPING << relativePos << COMMA;
						if (reverseIter->cigar != NULL) {
							resultStream << reverseIter->cigar->str() << COMMA;
						} else {
							resultStream << gencigar << SEPARATOR;
						}
						resultStream << reverseIter->distance << SEMI_COLON;
					}
					resultStream << endl;
				} else if (reverseIter != reverseEnd) {
					// Write the first result.
					samFlags += READ_PAIR_SECOND_REVERSE_MAPPED;
					getChromosomeOffset(searcher->indexer->chromosomeNames,
							searcher->indexer->chromosomeLengths, reverseIter->pos, chrName,
							relativePos);
					relativePos++;
					resultStream << concreteReadSt->second.readId << SEPARATOR << samFlags
							<< SEPARATOR << chrName << SEPARATOR << relativePos << SEPARATOR;
					resultStream << mapQuality << SEPARATOR;
					if (reverseIter->cigar != NULL) {
						resultStream << reverseIter->cigar->str() << SEPARATOR;
					} else {
						resultStream << gencigar << SEPARATOR;
					}
					resultStream << refNext << SEPARATOR;
					resultStream << templateLen << SEPARATOR
							<< concreteReadSt->first.reverseReadString << SEPARATOR
							<< concreteReadSt->first.reverseReadQuality << SEPARATOR;
					resultStream << NM_TAG << reverseIter->distance << SEPARATOR;
					reverseIter++;
					if (reverseIter != reverseEnd) {
						resultStream << SEPARATOR << XA_Z_TAG;
					}
					for (; reverseIter != reverseEnd; reverseIter++) {
						getChromosomeOffset(searcher->indexer->chromosomeNames,
								searcher->indexer->chromosomeLengths, reverseIter->pos, chrName,
								relativePos);
						relativePos++;
						resultStream << chrName << COMMA << REV_MAPPING << relativePos << COMMA;
						if (reverseIter->cigar != NULL) {
							resultStream << reverseIter->cigar->str() << COMMA;
						} else {
							resultStream << gencigar << SEPARATOR;
						}
						resultStream << reverseIter->distance << SEMI_COLON;
					}
					resultStream << endl;
				}
			} else {
				// Second read no mapping.
				samFlags = READ_PAIR_SECOND_UNMAPPED;
				resultStream << concreteReadSt->second.readId << SEPARATOR << samFlags << SEPARATOR
						<< UNAVAILABLE_1 << SEPARATOR << UNAVAILABLE_2 << SEPARATOR << UNAVAILABLE_2
						<< SEPARATOR << UNAVAILABLE_1 << SEPARATOR << UNAVAILABLE_1 << SEPARATOR
						<< UNAVAILABLE_2 << SEPARATOR << UNAVAILABLE_2 << SEPARATOR
						<< concreteReadSt->second.readString << SEPARATOR
						<< concreteReadSt->second.readQuality << endl;
			}
		}
	}
}

void DebugFormatResultStringWriter::writeResultString(stringstream& resultStream,
		ReadInfo& readInfo, Searcher* searcher) {
	SingleEndSearcher* concreteSearcher = static_cast<SingleEndSearcher*>(searcher);
	resultStream << readInfo.readId << " " << concreteSearcher->forwardResults.size() << " "
			<< concreteSearcher->reverseResults.size() << endl;
}

void PairedDebugFormatResultStringWriter::writeResultString(stringstream& resultStream,
		ReadInfo& readInfo, Searcher* searcher) {
	PairedEndSearcher* concreteSearcher = static_cast<PairedEndSearcher*>(searcher);
	PairedEndReadInfo* concreteReadSt = static_cast<PairedEndReadInfo*>(&readInfo);
	resultStream << concreteReadSt->first.readId << "\t" << concreteSearcher->forwardResults.size()
			<< "\t" << concreteSearcher->reverseResults.size() << endl;
}

void PairedEnhancedDebugFormatResultStringWriter::writeResultString(stringstream& resultStream,
		ReadInfo& readInfo, Searcher* searcher) {
	PairedEndSearcherEnhanced* concreteSearcher = dynamic_cast<PairedEndSearcherEnhanced*>(searcher);
	PairedEndReadInfo* concreteReadSt = static_cast<PairedEndReadInfo*>(&readInfo);
	if (concreteSearcher->forwardResults.size() != 0
			|| concreteSearcher->reverseResults.size() != 0) {
		resultStream << concreteReadSt->first.readId << "\t"
				<< concreteSearcher->forwardResults.size() << "\t"
				<< concreteSearcher->reverseResults.size() << endl;
	} else {
		if (concreteSearcher->individualForwardResults1.size() != 0
				|| concreteSearcher->individualReverseResults2.size() != 0) {
			resultStream << concreteReadSt->first.readId << "\t"
					<< concreteSearcher->individualForwardResults1.size() << "\t"
					<< concreteSearcher->individualReverseResults2.size() << endl;
		}
		if (concreteSearcher->individualForwardResults2.size() != 0
				|| concreteSearcher->individualReverseResults1.size() != 0) {
			resultStream << concreteReadSt->second.readId << "\t"
					<< concreteSearcher->individualForwardResults2.size() << "\t"
					<< concreteSearcher->individualReverseResults1.size() << endl;
		}
	}
}
