// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "search/pairedendverifier.h"

#include <algorithm>
#include <limits>

using namespace std;

void PairedEndVerifier::setMinMaxDist(unsigned readLen) {
	if (minInsertSize < readLen) {
		minInsertSize = readLen;
	}
	minDist = minInsertSize - readLen;
	maxDist = maxInsertSize - readLen;
}

void PairedEndVerifier::verifyCandidates(PairedEndRead& read, vector<unsigned>& candidates1,
		vector<unsigned>& candidates2, vector<PairedReadMapping>& results) {
	const string& read1 = *read.first.readStr;
	const string& read2 = *read.second.readStr;
	// Assuming both reads have the same length.
	unsigned readLen = read1.size();
	// Going over the candidates.
	vector<unsigned>::const_iterator candSet1Iter = candidates1.begin();
	vector<unsigned>::const_iterator candSet2Iter = candidates2.begin();
	vector<unsigned>::const_iterator backup = candidates2.begin();
	vector<unsigned>::const_iterator candSet1End = candidates1.end();
	vector<unsigned>::const_iterator candSet2End = candidates2.end();
	while (candSet1Iter != candSet1End && candSet2Iter != candSet2End) {
		if (*candSet2Iter >= (*candSet1Iter + minDist)
				&& *candSet2Iter <= (*candSet1Iter + maxDist)) {
			unsigned currentCandidate1 = *candSet1Iter;
			unsigned currentCandidate2 = *candSet2Iter++;
			__builtin_prefetch((const void*) (refSeqPtr + (*(candSet1Iter + CACHE_PREFETCH_STEP))),
					0, 0);
			__builtin_prefetch((const void*) (refSeqPtr + (*(candSet2Iter + CACHE_PREFETCH_STEP))),
					0, 0);
			unsigned refMatchLen1 = -1;
			int offset1 = 0;
			int offset2 = 0;
			unsigned distance1 = singleEndVerifier->getDistance(refSeqPtr + currentCandidate1,
					readLen, read1, offset1, refMatchLen1);
			if (distance1 <= distThresh) {
				unsigned distance2 = singleEndVerifier->getDistance(refSeqPtr + currentCandidate2,
						readLen, read2, offset2, refMatchLen1);
				if (distance2 <= distThresh) {
					vector<unsigned>::const_iterator it = upper_bound(indexer->chromosomeLengths.begin(),
							indexer->chromosomeLengths.end(), currentCandidate1);
					if (currentCandidate2 + readLen <= *it) {
						tempMisMatch[distance1 + distance2]++;
						results.push_back(
								PairedReadMapping(currentCandidate1 + offset1, distance1,
										currentCandidate2 + offset2, distance2, NULL, NULL));
					}
				}
			}
		}
		while (candSet1Iter != candSet1End && *candSet2Iter > (*candSet1Iter + maxDist)) {
			candSet1Iter++;
			candSet2Iter = backup;
		}
		while (candSet2Iter != candSet2End && candSet1Iter != candSet1End
				&& *candSet2Iter < (*candSet1Iter + minDist)) {
			candSet2Iter++;
			backup++;
		}
	}
}

void PairedEndIndelVerifier::verifyCandidates(PairedEndRead& read, vector<unsigned>& candidates1,
		vector<unsigned>& candidates2, vector<PairedReadMapping>& results) {
	SingleEndIndelVerifier* indelVerifier = static_cast<SingleEndIndelVerifier*>(singleEndVerifier);

	const string& read1 = *read.first.readStr;
	const string& read2 = *read.second.readStr;
	// Assuming both reads have the same length.
	unsigned readLen = read1.size();
	// Going over the candidates.
	vector<unsigned>::const_iterator candSet1Iter = candidates1.begin();
	vector<unsigned>::const_iterator candSet2Iter = candidates2.begin();
	vector<unsigned>::const_iterator backup = candidates2.begin();
	vector<unsigned>::const_iterator candSet1End = candidates1.end();
	vector<unsigned>::const_iterator candSet2End = candidates2.end();
	GramCountArray readGramCounts1;
	GramCountArray readGramCounts2;
	indelVerifier->fillGramCountArray(read1.data(), read1.length(), readGramCounts1);
	indelVerifier->fillGramCountArray(read2.data(), read2.length(), readGramCounts2);
	unsigned distance1 = numeric_limits<unsigned>::max();
	unsigned distance2 = numeric_limits<unsigned>::max();
	while (candSet1Iter != candSet1End && candSet2Iter != candSet2End) {
		if (*candSet2Iter >= (*candSet1Iter + minDist)
				&& *candSet2Iter <= (*candSet1Iter + maxDist)) {
			// This is a valid paired candidate.
			unsigned currentCandidate1 = *candSet1Iter;
			unsigned currentCandidate2 = *candSet2Iter++;
			__builtin_prefetch((const void*) (refSeqPtr + (*(candSet1Iter + CACHE_PREFETCH_STEP))),
					0, 0);
			__builtin_prefetch((const void*) (refSeqPtr + (*(candSet2Iter + CACHE_PREFETCH_STEP))),
					0, 0);
			unsigned lowerBound = indelVerifier->getDistanceLowerBound(readGramCounts1,
					currentCandidate1, readLen);
			unsigned refMatchLen1 = -1;
			unsigned refMatchLen2 = -1;
			int flag=0;
			unsigned maxLength = readLen + distThresh;
			if (lowerBound <= (distThresh + 1)) {
				unsigned lowerBound2 = indelVerifier->getDistanceLowerBound(readGramCounts2,
						currentCandidate2, readLen);
				if (lowerBound2 <= (distThresh + 1)) {
					stringstream* cigar1 = NULL;
					stringstream* cigar2 = NULL;
					if(genCigar){
						unsigned offset1 = 0;
						cigar1 = new stringstream();
						unsigned indel1=-1;
						distance1= indelVerifier->getBitVectorEditDistance(refSeqPtr + currentCandidate1, refMatchLen1,maxLength,
							read1,distThresh, *cigar1,indel1, offset1);
						if (distance1 <= distThresh) {
							flag=1;
							cigar2 = new stringstream();
							unsigned indel2=-1;
							unsigned offset2 = 0;
							distance2= indelVerifier->getBitVectorEditDistance(refSeqPtr + currentCandidate2, refMatchLen2,maxLength,
								read2,distThresh, *cigar2,indel2, offset2);
							if (distance2 <= distThresh) {
								vector<unsigned>::const_iterator it = upper_bound(
										indexer->chromosomeLengths.begin(),
										indexer->chromosomeLengths.end(), currentCandidate1);
								vector<unsigned>::const_iterator it2 = upper_bound(
										indexer->chromosomeLengths.begin(),
										indexer->chromosomeLengths.end(), currentCandidate2);
								if ((currentCandidate1 + readLen <= *it)
										&& (currentCandidate2 + readLen <= *it2)) {
									string chrName = indexer->chromosomeNames.at(
											(it - indexer->chromosomeLengths.begin()));
									string chrName1 = indexer->chromosomeNames.at(
											(it2 - indexer->chromosomeLengths.begin()));
										if (chrName == chrName1) {
											flag=2;
											if (indel1 > 0 || indel2 > 0) {
												tempIndel[indel1 + indel2]++;
											}
											else {
												tempMisMatch[distance1 + distance2]++;
											}
											results.push_back(
													PairedReadMapping(currentCandidate1 + offset1,
															distance1, currentCandidate2 + offset2,
															distance2, cigar1, cigar2));
										}
									}
								}
							}
						else{
							delete cigar1;
						}
						if(flag==1){
							delete cigar1;
							delete cigar2;
						}

					}
					else{
						distance1= indelVerifier->getBitVectorEditDistance(refSeqPtr + currentCandidate1, refMatchLen1,maxLength,
							read1,distThresh);
						if (distance1 <= distThresh) {
							distance2= indelVerifier->getBitVectorEditDistance(refSeqPtr + currentCandidate2, refMatchLen2,maxLength,
								read2,distThresh);
							if (distance2 <= distThresh) {
								vector<unsigned>::const_iterator it = upper_bound(
										indexer->chromosomeLengths.begin(),
										indexer->chromosomeLengths.end(), currentCandidate1);
								vector<unsigned>::const_iterator it2 = upper_bound(
										indexer->chromosomeLengths.begin(),
										indexer->chromosomeLengths.end(), currentCandidate2);
								if ((currentCandidate1 + readLen <= *it)
										&& (currentCandidate2 + readLen <= *it2)) {
									string chrName = indexer->chromosomeNames.at(
											(it - indexer->chromosomeLengths.begin()));
									string chrName1 = indexer->chromosomeNames.at(
											(it2 - indexer->chromosomeLengths.begin()));
									if (chrName == chrName1) {
										tempMisMatch[distance1 + distance2]++;
										results.push_back(
												PairedReadMapping(currentCandidate1,
														distance1, currentCandidate2,
														distance2, cigar1, cigar2));
									}
								}
							}
						}
					}
				}
			}
		}
		while (candSet1Iter != candSet1End && *candSet2Iter > (*candSet1Iter + maxDist)) {
			candSet1Iter++;
			candSet2Iter = backup;
		}
		while (candSet2Iter != candSet2End && *candSet2Iter < (*candSet1Iter + minDist)) {
			candSet2Iter++;
			backup++;
		}
	}
}

void PairedEndVerifierEnhanced::verifyCandidates(PairedEndRead& read, vector<unsigned>& candidates1,
		vector<unsigned>& candidates2, vector<PairedReadMapping>& results,
		vector<ReadMapping>& results1, vector<ReadMapping>& results2) {
	PairedEndVerifier::verifyCandidates(read, candidates1, candidates2, results);
	if (results.size() != 0) {
		return;
	}
	singleEndVerifier->verifyCandidates(*read.first.readStr, candidates1, results1);
	singleEndVerifier->verifyCandidates(*read.second.readStr, candidates2, results2);
}
