// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "search/singleendmerger.h"

#include <iostream>
#include <cstdlib>
#include <limits>

using namespace std;

void SingleEndIndelMerger::findCandidatesAll(Read &read, vector<unsigned>& candidates) {
	const vector<BitVectorInvertedListWithOffset>& invLists = read.bitvectorInvertedLists;
	const vector<unsigned>& prefixGrams = read.prefixGrams;
	// Panic case. Return nothing.
	if ((invLists.size() < distThresh) || (distThresh == 0)) {
		return;
	}
	vector<BitVecPair>::const_iterator invListIter = invLists[prefixGrams[0]].second->begin();
	vector<BitVecPair>::const_iterator invListEnd = invLists[prefixGrams[0]].second->end();
	unsigned gramPosition = invLists[prefixGrams[0]].first;
	tmp.clear();
	while (invListIter != invListEnd) {
		if (invListIter->pos >= gramPosition) {
			break;
		}
		invListIter++;
	}
	while (invListIter != invListEnd) {
		candidates.push_back((invListIter->pos - gramPosition));
		invListIter++;
	}
	for (unsigned j = 1; j < prefixGrams.size(); j++) {
		unsigned i = prefixGrams[j];
		unsigned gramPosition = invLists[i].first;
		tmp.clear();
		tmp.reserve(candidates.size() + invLists[i].second->size());
		vector<unsigned>::iterator candIter = candidates.begin();
		invListIter = invLists[i].second->begin();
		invListEnd = invLists[i].second->end();
		while (invListIter != invListEnd) {
			if (invListIter->pos >= gramPosition) {
				break;
			}
			invListIter++;
		}
		while (candIter != candidates.end()) {
			if ((invListIter != invListEnd && *candIter > (invListIter->pos - gramPosition))) {
				tmp.push_back((invListIter->pos - gramPosition));
				invListIter++;
			} else if (invListIter == invListEnd
					|| (*candIter < (invListIter->pos - gramPosition))) {
				tmp.push_back(*candIter++);
			} else {
				tmp.push_back(*candIter++);
				invListIter++;
			}
		}
		while (invListIter < invListEnd) {
			tmp.push_back((invListIter->pos - gramPosition));
			invListIter++;
		}
		std::swap(candidates, tmp);
	}
}

bool SingleEndIndelMerger::findCandidatesAndExactMappingsK(unsigned& k, Read& read,
		vector<unsigned> &candidates, vector<ReadMapping> &results,
		SingleEndVerifier& verifier) {
	cerr << "Error: K mappings mode not supported for edit distance." << endl;
	exit(0);
}

bool SingleEndIndelMerger::findCandidatesAndMappingsK(unsigned& k, Read& read,
		vector<unsigned> &candidates, vector<ReadMapping> &results,
		SingleEndVerifier& verifier) {
	cerr << "Error: K mappings mode not supported for edit distance." << endl;
	exit(0);
}

void SingleEndHeuristicIndelMerger::findCandidatesAll(Read &read, vector<unsigned>& candidates) {
	const vector<BitVectorInvertedListWithOffset>& invLists = read.bitvectorInvertedLists;
	const vector<unsigned>& prefixGrams = read.prefixGrams;
	const string& readStr = *read.readStr;
	vector<BitVecPair>::const_iterator invListIter = invLists[prefixGrams[0]].second->begin();
	vector<BitVecPair>::const_iterator invListEnd = invLists[prefixGrams[0]].second->end();
	unsigned gramPosition = invLists[prefixGrams[0]].first;
	BitvectorType alignMask = numeric_limits<BitvectorType>::max();
	BitvectorType readBitVectorPerGram = generateBitVector(alignMask, readStr, gramPosition,
			indexer->gramGen.gramLen, readStr.size());
	while (invListIter != invListEnd) {
		if (invListIter->pos >= gramPosition) {
			break;
		}
		invListIter++;
	}
	while (invListIter != invListEnd) {
		if (GCC_POPCOUNT(readBitVectorPerGram ^ (invListIter->bitVec & alignMask))
				<= bitVectorThresh) {
			candidates.push_back((invListIter->pos - gramPosition));
		}
		invListIter++;
	}
	for (unsigned j = 1; j < prefixGrams.size(); j++) {
		unsigned i = prefixGrams[j];
		unsigned gramPosition = invLists[i].first;
		BitvectorType alignMask = numeric_limits<BitvectorType>::max();
		BitvectorType readBitVectorPerGram = generateBitVector(alignMask, readStr, gramPosition,
				indexer->gramGen.gramLen, readStr.size());
		tmp.clear();
		tmp.reserve(candidates.size() + invLists[i].second->size());
		vector<unsigned>::const_iterator candIter = candidates.begin();
		invListIter = invLists[i].second->begin();
		invListEnd = invLists[i].second->end();
		while (invListIter != invListEnd) {
			if (invListIter->pos >= gramPosition) {
				break;
			}
			invListIter++;
		}
		while (candIter != candidates.end()) {
			if ((invListIter != invListEnd && *candIter > (invListIter->pos - gramPosition))) {
				if (GCC_POPCOUNT(readBitVectorPerGram ^ (invListIter->bitVec & alignMask))
						<= bitVectorThresh) {
					tmp.push_back((invListIter->pos - gramPosition));
				}
				invListIter++;
			} else if (invListIter == invListEnd
					|| (*candIter < (invListIter->pos - gramPosition))) {
				tmp.push_back(*candIter++);
			} else {
				tmp.push_back(*candIter++);
				invListIter++;
			}
		}
		while (invListIter < invListEnd) {
			if (GCC_POPCOUNT(readBitVectorPerGram ^ (invListIter->bitVec & alignMask))
					<= bitVectorThresh) {
				tmp.push_back((invListIter->pos - gramPosition));
			}

			invListIter++;
		}
		std::swap(candidates, tmp);
	}
}


bool SingleEndHeuristicIndelMerger::findCandidatesAndExactMappingsK(unsigned& k, Read& read,
		vector<unsigned>& candidates, vector<ReadMapping>& results, SingleEndVerifier& verifier) {
	unsigned kCount = k;
	unsigned loopCount = 0;
	vector<BitVectorInvertedListWithOffset>& invLists = read.bitvectorInvertedLists;
	vector<unsigned>& prefixGrams = read.prefixGrams;
	vector<BitVecPair>::const_iterator invListIter = invLists[prefixGrams[0]].second->begin();
	vector<BitVecPair>::const_iterator invListEnd = invLists[prefixGrams[0]].second->end();
	unsigned gramPosition = invLists[prefixGrams[0]].first;
	BitvectorType alignMask = numeric_limits<BitvectorType>::max();
	BitvectorType readBitVectorPerGram = generateBitVector(alignMask, *read.readStr, gramPosition,
			indexer->gramGen.gramLen, read.readStr->size());
	while (invListIter != invListEnd) {
		if (invListIter->pos >= gramPosition) {
			break;
		}
		invListIter++;
	}
	while (invListIter != invListEnd) {
		if (GCC_POPCOUNT(readBitVectorPerGram ^ (invListIter->bitVec & alignMask))
				<= bitVectorThresh) {
			candidates.push_back((invListIter->pos - gramPosition));
			if ((candidates.size() % kCount) == 0) {
				verifier.verifyCandidatesExact(loopCount, k, *read.readStr, candidates, results);
				if (k == 0) {
					return true;
				}
				loopCount += kCount;
			}
		}
		invListIter++;
	}
	if ((candidates.size() % kCount) != 0) {
		verifier.verifyCandidatesExact(loopCount, k, *read.readStr, candidates, results);
		if (k == 0) {
			return true;
		}
	}
	return false;
}

bool SingleEndHeuristicIndelMerger::findCandidatesAndMappingsK(unsigned& k, Read& read,
		vector<unsigned>& candidates, vector<ReadMapping>& results, SingleEndVerifier& verifier) {
	verifier.distThresh = distThresh;
	unsigned kCount = k;
	unsigned loopCount = 0;
	vector<BitVectorInvertedListWithOffset>& bitVectorInvertedLists = read.bitvectorInvertedLists;
	vector<unsigned>& prefixGrams = read.prefixGrams;
	vector<BitVecPair>::const_iterator invListIter;
	vector<BitVecPair>::const_iterator invListEnd;
	tmp.clear();
	tmp1.clear();
	tmp1.reserve(kCount + 1);
	verifier.verifyCandidates(loopCount, k, *read.readStr, candidates, results);
	if (k == 0) {
		return true;
	}
	for (unsigned j = 1; j < prefixGrams.size(); j++) {
		unsigned i = prefixGrams[j];
		unsigned gramPosition = bitVectorInvertedLists[i].first;
		BitvectorType alignMask = numeric_limits<BitvectorType>::max();
		BitvectorType readBitVectorPerGram = generateBitVector(alignMask, *read.readStr,
				gramPosition, indexer->gramGen.gramLen, read.readStr->size());
		tmp.clear();
		tmp.reserve(candidates.size() + bitVectorInvertedLists[i].second->size());
		vector<unsigned>::const_iterator candIter = candidates.begin();
		invListIter = bitVectorInvertedLists[i].second->begin();
		invListEnd = bitVectorInvertedLists[i].second->end();
		while (invListIter != invListEnd) {
			if (invListIter->pos >= gramPosition) {
				break;
			}
			invListIter++;
		}
		while (candIter != candidates.end()) {
			if ((invListIter != invListEnd && *candIter > (invListIter->pos - gramPosition))) {
				if (GCC_POPCOUNT(readBitVectorPerGram ^ (invListIter->bitVec & alignMask))
						<= bitVectorThresh) {
					tmp.push_back((invListIter->pos - gramPosition));
					tmp1.push_back((invListIter->pos - gramPosition));
					if ((tmp1.size() % kCount) == 0) {
						verifier.verifyCandidates(loopCount, k, *read.readStr, tmp1, results);
						tmp1.clear();
						if (k == 0) {
							return true;
						}
					}
				}
				invListIter++;
			} else if (invListIter == invListEnd
					|| (*candIter < (invListIter->pos - gramPosition))) {
				tmp.push_back(*candIter++);
			} else {
				tmp.push_back(*candIter);
				candIter++;
				invListIter++;
			}
		}
		while (invListIter < invListEnd) {
			if (GCC_POPCOUNT(readBitVectorPerGram ^ (invListIter->bitVec & alignMask))
					<= bitVectorThresh) {
				tmp.push_back((invListIter->pos - gramPosition));
				tmp1.push_back((invListIter->pos - gramPosition));
				if ((tmp1.size() % kCount) == 0) {
					verifier.verifyCandidates(loopCount, k, *read.readStr, tmp1, results);
					tmp1.clear();
					if (k == 0) {
						return true;
					}
				}
			}
			invListIter++;
		}
		if (tmp1.size() != 0) {
			verifier.verifyCandidates(loopCount, k, *read.readStr, tmp1, results);
			tmp1.clear();
			if (k == 0) {
				return true;
			}
		}
		std::swap(candidates, tmp);
	}
	return false;
}

void SingleEndHammingMerger::findCandidatesAll(Read& read, vector<unsigned>& candidates) {
	const vector<BitVectorInvertedListWithOffset>& invLists = read.bitvectorInvertedLists;
	const vector<unsigned>& prefixGrams = read.prefixGrams;
	const string& readStr = *read.readStr;
	vector<BitVecPair>::const_iterator invListIter = invLists[prefixGrams[0]].second->begin();
	vector<BitVecPair>::const_iterator invListEnd = invLists[prefixGrams[0]].second->end();
	unsigned gramPosition = invLists[prefixGrams[0]].first;
	BitvectorType alignMask = numeric_limits<BitvectorType>::max();
	BitvectorType readBitVectorPerGram = generateBitVector(alignMask, readStr, gramPosition,
			indexer->gramGen.gramLen, readStr.size());
	while (invListIter != invListEnd) {
		if (invListIter->pos >= gramPosition) {
			break;
		}
		invListIter++;
	}
	while (invListIter != invListEnd) {
		if (GCC_POPCOUNT(readBitVectorPerGram ^ (invListIter->bitVec & alignMask)) <= distThresh) {
			candidates.push_back((invListIter->pos - gramPosition));
		}
		invListIter++;
	}
	for (unsigned j = 1; j < prefixGrams.size(); j++) {
		unsigned i = prefixGrams[j];
		unsigned gramPosition = invLists[i].first;
		BitvectorType alignMask = numeric_limits<BitvectorType>::max();
		BitvectorType readBitVectorPerGram = generateBitVector(alignMask, readStr, gramPosition,
				indexer->gramGen.gramLen, readStr.size());
		tmp.clear();
		tmp.reserve(candidates.size() + invLists[i].second->size());
		vector<unsigned>::const_iterator candIter = candidates.begin();
		invListIter = invLists[i].second->begin();
		invListEnd = invLists[i].second->end();
		while (invListIter != invListEnd) {
			if (invListIter->pos >= gramPosition) {
				break;
			}
			invListIter++;
		}
		while (candIter != candidates.end()) {
			if ((invListIter != invListEnd && *candIter > (invListIter->pos - gramPosition))) {
				if (GCC_POPCOUNT(readBitVectorPerGram ^ (invListIter->bitVec & alignMask))
						<= distThresh) {
					tmp.push_back((invListIter->pos - gramPosition));
				}
				invListIter++;
			} else if (invListIter == invListEnd
					|| (*candIter < (invListIter->pos - gramPosition))) {
				tmp.push_back(*candIter++);
			} else {
				if (GCC_POPCOUNT(readBitVectorPerGram ^ (invListIter->bitVec & alignMask))
						<= distThresh) {
					tmp.push_back(*candIter);
				}
				candIter++;
				invListIter++;
			}
		}
		while (invListIter < invListEnd) {
			if (GCC_POPCOUNT(readBitVectorPerGram ^ (invListIter->bitVec & alignMask))
					<= distThresh) {
				tmp.push_back((invListIter->pos - gramPosition));
			}
			invListIter++;
		}
		std::swap(candidates, tmp);
	}
}

bool SingleEndHammingMerger::findCandidatesAndExactMappingsK(unsigned& k, Read& read,
		vector<unsigned>& candidates, vector<ReadMapping>& results, SingleEndVerifier& verifier) {
	verifier.distThresh = 0;
	unsigned kCount = k;
	unsigned loopCount = 0;
	vector<BitVectorInvertedListWithOffset>& invLists = read.bitvectorInvertedLists;
	vector<unsigned>& prefixGrams = read.prefixGrams;
	vector<BitVecPair>::const_iterator invListIter = invLists[prefixGrams[0]].second->begin();
	vector<BitVecPair>::const_iterator invListEnd = invLists[prefixGrams[0]].second->end();
	unsigned gramPosition = invLists[prefixGrams[0]].first;
	BitvectorType alignMask = numeric_limits<BitvectorType>::max();
	BitvectorType readBitVectorPerGram = generateBitVector(alignMask, *read.readStr, gramPosition,
			indexer->gramGen.gramLen, read.readStr->size());
	while (invListIter != invListEnd) {
		if (invListIter->pos >= gramPosition) {
			break;
		}
		invListIter++;
	}
	while (invListIter != invListEnd) {
		if (GCC_POPCOUNT(readBitVectorPerGram ^ (invListIter->bitVec & alignMask)) <= distThresh) {
			candidates.push_back((invListIter->pos - gramPosition));
			if ((candidates.size() % kCount) == 0) {
				verifier.verifyCandidatesExact(loopCount, k, *read.readStr, candidates, results);
				if (k == 0) {
					return true;
				}
				loopCount += kCount;
			}
		}
		invListIter++;
	}
	if ((candidates.size() % kCount) != 0) {
		verifier.verifyCandidatesExact(loopCount, k, *read.readStr, candidates, results);
		if (k == 0) {
			return true;
		}
	}
	return false;
}

bool SingleEndHammingMerger::findCandidatesAndMappingsK(unsigned& k, Read& read,
		vector<unsigned>& candidates, vector<ReadMapping>& results, SingleEndVerifier& verifier) {
	verifier.distThresh = distThresh;
	unsigned kCount = k;
	unsigned loopCount = 0;
	vector<BitVectorInvertedListWithOffset>& invLists = read.bitvectorInvertedLists;
	vector<unsigned>& prefixGrams = read.prefixGrams;
	vector<BitVecPair>::const_iterator invListIter;
	vector<BitVecPair>::const_iterator invListEnd;
	tmp.clear();
	tmp1.clear();
	tmp1.reserve(kCount + 1);
	verifier.verifyCandidates(loopCount, k, *read.readStr, candidates, results);
	if (k == 0) {
		return true;
	}
	for (unsigned j = 1; j < prefixGrams.size(); j++) {
		unsigned i = prefixGrams[j];
		unsigned gramPosition = invLists[i].first;
		BitvectorType alignMask = numeric_limits<BitvectorType>::max();
		BitvectorType readBitVectorPerGram = generateBitVector(alignMask, *read.readStr,
				gramPosition, indexer->gramGen.gramLen, read.readStr->size());
		tmp.clear();
		tmp.reserve(candidates.size() + invLists[i].second->size());
		vector<unsigned>::const_iterator candIter = candidates.begin();
		invListIter = invLists[i].second->begin();
		invListEnd = invLists[i].second->end();
		while (invListIter != invListEnd) {
			if (invListIter->pos >= gramPosition) {
				break;
			}
			invListIter++;
		}
		while (candIter != candidates.end()) {
			if ((invListIter != invListEnd && *candIter > (invListIter->pos - gramPosition))) {
				if (GCC_POPCOUNT(readBitVectorPerGram ^ (invListIter->bitVec & alignMask))
						<= distThresh) {
					tmp.push_back((invListIter->pos - gramPosition));
					tmp1.push_back((invListIter->pos - gramPosition));
					if ((tmp1.size() % kCount) == 0) {
						verifier.verifyCandidates(loopCount, k, *read.readStr, tmp1, results);
						tmp1.clear();
						if (k == 0) {
							return true;
						}
					}
				}
				invListIter++;
			} else if (invListIter == invListEnd
					|| (*candIter < (invListIter->pos - gramPosition))) {
				tmp.push_back(*candIter++);
			} else {
				if (GCC_POPCOUNT(readBitVectorPerGram ^ (invListIter->bitVec & alignMask))
						<= distThresh) {
					tmp.push_back(*candIter);
				}
				candIter++;
				invListIter++;
			}
		}
		while (invListIter < invListEnd) {
			if (GCC_POPCOUNT(readBitVectorPerGram ^ (invListIter->bitVec & alignMask))
					<= distThresh) {
				tmp.push_back((invListIter->pos - gramPosition));
				tmp1.push_back((invListIter->pos - gramPosition));
				if ((tmp1.size() % kCount) == 0) {
					verifier.verifyCandidates(loopCount, k, *read.readStr, tmp1, results);
					tmp1.clear();
					if (k == 0) {
						return true;
					}
				}
			}
			invListIter++;
		}
		if (tmp1.size() != 0) {
			verifier.verifyCandidates(loopCount, k, *read.readStr, tmp1, results);
			tmp1.clear();
			if (k == 0) {
				return true;
			}
		}
		std::swap(candidates, tmp);
	}
	return false;
}
