// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "include/globals.h"
#include "include/bitvectorfilter.h"
#include "include/search/verifier.h"
#include "include/search/singleendverifier.h"
#include "include/search/pairedendverifier.h"
#include "include/search/bitvector.h"
#include <seqan/align.h>
#include <limits>

using namespace std;
using namespace seqan;

void SingleEndVerifier::verifyCandidatesExact(unsigned& startPos, unsigned& k, const string& read,
		vector<unsigned>& candidates, vector<ReadMapping>& results) {
	unsigned readLen = read.size();
	unsigned* candi = candidates.data() + startPos;
	unsigned* candiEnd = candidates.data() + candidates.size();
	while (candi != candiEnd) {
		unsigned currentCandidate = *candi++;
		__builtin_prefetch((const void*) (refSeqPtr + (*(candi + CACHE_PREFETCH_STEP))), 0, 0);
		if (exactMatch(refSeqPtr + currentCandidate, readLen, read)) {
			vector<unsigned>::iterator it = upper_bound(indexer->chromosomeLengths.begin(),
					indexer->chromosomeLengths.end(), currentCandidate);
			if (currentCandidate + readLen <= *it) {
				tempMisMatch[0]++;
				results.push_back(ReadMapping(currentCandidate, 0, NULL));
				k--;
			}
		}
		if (k == 0) {
			return;
		}
	}
}

bool SingleEndVerifier::exactMatch(const char* candi, unsigned candiLen, const std::string& read) {
	const char* s2 = read.data();
	const long* ucandi = (const long*) candi;
	const long* uquery = (const long*) s2;
	unsigned ppNumChunks = candiLen / sizeof(long);
	// Check for equality in long chunks.
	for (unsigned i = 0; i < ppNumChunks; i++) {
		if (*ucandi++ != *uquery++) {
			return false;
		}
	}
	// Check remaining characters.
	unsigned remainingChars = candiLen - ppNumChunks * sizeof(long);
	if (remainingChars > 0) {
		const char* candiRemain = candi + ppNumChunks * sizeof(long);
		const char* readRemain = read.data() + ppNumChunks * sizeof(long);
		for (unsigned i = 0; i < remainingChars; i++) {
			if (*candiRemain++ != *readRemain++) {
				return false;
			}
		}
	}
	return true;
}

void SingleEndHammingVerifier::verifyCandidates(const string& read, vector<unsigned>& candidates,
		vector<ReadMapping>& results) {
	unsigned readLen = read.size();
	unsigned* candi = candidates.data();
	unsigned* candiEnd = candidates.data() + candidates.size();
	while (candi != candiEnd) {
		unsigned currentCandidate = *candi++;
		__builtin_prefetch((const void*) (refSeqPtr + (*(candi + CACHE_PREFETCH_STEP))), 0, 0);
		unsigned distance = getDistance(refSeqPtr + currentCandidate, readLen, read);
		if (distance <= distThresh) {
			vector<unsigned>::iterator it = upper_bound(indexer->chromosomeLengths.begin(),
					indexer->chromosomeLengths.end(), currentCandidate);
			if (currentCandidate + readLen <= *it) {
				tempMisMatch[distance]++;
				results.push_back(ReadMapping(currentCandidate, distance, NULL));
			}
		}
	}
}

// evaluates mismatches only - or hamming distance

void SingleEndHammingVerifier::verifyCandidates(unsigned& startPos, unsigned& k, const string& read,
		vector<unsigned>& candidates, vector<ReadMapping>& results) {
	unsigned readLen = read.size();
	unsigned* candi = candidates.data() + startPos;
	unsigned* candiEnd = candidates.data() + candidates.size();
	while (candi != candiEnd) {
		unsigned currentCandidate = *candi++;
		__builtin_prefetch((const void*) (refSeqPtr + (*(candi + CACHE_PREFETCH_STEP))), 0, 0);
		unsigned distance = getDistance(refSeqPtr + currentCandidate, readLen, read);
		// Thanh T. Nguyen
		// too many comparisons to mean something very simple
		//if (distance <= distThresh) {
		if ((distThresh == 0 && distance == 0) || (distance <= distThresh && distance > 0)) {
			vector<unsigned>::iterator it = upper_bound(indexer->chromosomeLengths.begin(),
					indexer->chromosomeLengths.end(), currentCandidate);
			if (currentCandidate + readLen <= *it) {
				tempMisMatch[distance]++;
				results.push_back(ReadMapping(currentCandidate, distance, NULL));
				k--;
			}
		}
		if (k == 0) {
			return;
		}
	}
}

unsigned SingleEndHammingVerifier::getDistance(const char* candi, unsigned candiLen, const string& read) {
	unsigned dist = 0;
	const char* s2 = read.data();
	const PPFilterChunkType* ucandi = (const PPFilterChunkType*) candi;
	const PPFilterChunkType* uquery = (const PPFilterChunkType*) s2;
#ifdef READLEN
	unsigned ppNumChunks = READLEN / sizeof(PPFilterChunkType);
#else
	unsigned ppNumChunks = candiLen / sizeof(PPFilterChunkType);
#endif
	// Calculate hamming distance of chunks as a filtering step.
	for (unsigned i = 0; i < ppNumChunks; i++) {
		if (*ucandi++ != *uquery++) {
			dist++;
		}
	}
	if (dist > distThresh) {
		return dist;
	}
	dist = 0;

	// Calculate true hamming distance character-by-character.
	// Thanh T. Nguyen
	// - why are we checking every characters when we already have some matches.

	for (unsigned i = 0;
#ifdef READLEN
			i < READLEN;
#else
			i < candiLen;
#endif
			i++) {
		if (*candi++ != *s2++) {
			dist++;
		}
	}
	return dist;
}

unsigned SingleEndHammingVerifier::getDistance(const char* candi, unsigned candiLen, const string& read,
		int& offset, unsigned& refend) {
	return getDistance(candi, candiLen, read);
}

void SingleEndIndelVerifier::verifyCandidates(const string& read, vector<unsigned>& candidates,
		vector<ReadMapping>& results) {
	unsigned bestAnswerSoFar;
	unsigned lastChecked = 0;
	GramCountArray readGramCounts;
	fillGramCountArray(read.data(), read.length(), readGramCounts);
	unsigned readLen = read.size();
	unsigned* candi = candidates.data();
	unsigned* candiEnd = candidates.data() + candidates.size();
	unsigned dist = numeric_limits<unsigned>::max();

	while (candi != candiEnd) {

		unsigned currentCandidate = *candi++;
		unsigned resultForThisCandidate = 0;
		bestAnswerSoFar = distThresh + 1;
		unsigned refMatchLen = -1;
		unsigned start = (currentCandidate) > distThresh ? (currentCandidate) - distThresh : 0;
		start = max(start, lastChecked);
		unsigned end = currentCandidate + distThresh;
		if (currentCandidate >= start) {
			unsigned maxLength =
					((currentCandidate) + readLen + distThresh) < refSeq.length() ?
							(readLen + distThresh) : (refSeq.length() - (currentCandidate));
			unsigned lowerBound = getDistanceLowerBound(readGramCounts, currentCandidate, readLen);
			if (lowerBound <= distThresh) {
				dist = getBitVectorEditDistance(refSeqPtr + currentCandidate, refMatchLen,
						maxLength, read, distThresh);
				if (dist < bestAnswerSoFar) {
					bestAnswerSoFar = dist;
					resultForThisCandidate = currentCandidate;
				}
			}
		}
		if (bestAnswerSoFar == 0) {
			lastChecked = end;
			stringstream* cigar = NULL;
			if (genCigar) {
				cigar = new stringstream();
				unsigned indel = getCigar(refSeqPtr + resultForThisCandidate, read, *cigar,
						refMatchLen, bestAnswerSoFar);
				if (indel > 0) {
					tempIndel[indel]++;
				} else {
					tempMisMatch[bestAnswerSoFar]++;
				}
			} else {
				tempMisMatch[bestAnswerSoFar]++;
				results.push_back(ReadMapping(resultForThisCandidate, bestAnswerSoFar, cigar));
			}
			continue;
		} else {
			for (unsigned loc = currentCandidate - 1; loc >= start; loc--) {
				unsigned maxLength =
						((loc) + readLen + distThresh) < refSeq.length() ?
								(readLen + distThresh) : (refSeq.length() - (loc));
				unsigned lowerBound = getDistanceLowerBound(readGramCounts, loc, readLen);
				if (lowerBound <= distThresh) {
					if (bestAnswerSoFar > 0) {
						dist = getBitVectorEditDistance(refSeqPtr + loc, refMatchLen, maxLength,
								read, bestAnswerSoFar - 1);
					} else {
						dist = 0;
					}
					if (dist < bestAnswerSoFar) {
						bestAnswerSoFar = dist;
						resultForThisCandidate = loc;
					}
				}
			}
			for (unsigned loc = max(currentCandidate, lastChecked) + 1; loc <= end; loc++) {
				unsigned maxLength =
						((loc) + readLen + distThresh) < refSeq.length() ?
								(readLen + distThresh) : (refSeq.length() - (loc));
				unsigned lowerBound = getDistanceLowerBound(readGramCounts, loc, readLen);
				if (lowerBound <= distThresh) {
					if (bestAnswerSoFar > 0) {
						dist = getBitVectorEditDistance(refSeqPtr + loc, refMatchLen, maxLength,
								read, bestAnswerSoFar - 1);
                        cout << "ED1 = " << dist << endl;
					} else {
						dist = 0;
					}
					if (dist < bestAnswerSoFar) {
						bestAnswerSoFar = dist;
						resultForThisCandidate = loc;
					}
				}
			}
			lastChecked = end;
		}
		if (bestAnswerSoFar > distThresh) {
			continue;
		}
		stringstream* cigar = NULL;
		if (genCigar) {
			cigar = new stringstream();

			unsigned indel = getCigar(refSeqPtr + resultForThisCandidate, read, *cigar, refMatchLen,
					bestAnswerSoFar);

			if (indel > 0) {
				tempIndel[indel]++;
			} else {
				tempMisMatch[bestAnswerSoFar]++;
			}
		} else {
			tempMisMatch[bestAnswerSoFar]++;
			results.push_back(ReadMapping(resultForThisCandidate, bestAnswerSoFar, cigar));
		}
	}
}

int SingleEndIndelVerifier::BandedAlign(int allowed_errors, char * sref, char * sread, int reflen, int redlen, int z)
{
	//----------------------------------------------
	// first letter is treated separately
	//cout << endl;
	register int allowed_errors_plus_one = allowed_errors+1;
	register int i, w, max, j, d, t, l, tmax, iz,temp,mi,r_index;

	scores0 = &scores1D0[0];
	scores1 = &scores1D1[0];

	for (i=0;i<allowed_errors_plus_one;i++)
		scores0[i] = (sread[0]!=sref[i]);
	//----------------------------------------------
	// letters within the allowed errors
	w = 0;
	for (i=1;i<allowed_errors_plus_one;i++)
	{
		scores1[0] = (sread[i]!=sref[0]);
		if (scores0[0] > w)
			scores1[0] += i;
		else scores1[0] = scores0[0]+1;

		max = allowed_errors+i;
		for (j=1;j < max;j++)
		{
			scores1[j] = (sread[i]!=sref[j]);

			d = scores0[j-1];
			t = scores0[j];
			l = scores1[j-1];
			if (t < d)
			{
				if (t <= l)
					scores1[j] = t+1;
				else scores1[j] = l+1;
			} else if (l < d)
				scores1[j] = l+1;
			else scores1[j] += d;
		}
		tmax = max-1;
		scores1[max] = (sread[i]!=sref[max]);
		if (scores1[tmax] < scores0[tmax])
			scores1[max] = scores1[tmax]+1;
		else scores1[max] += scores0[tmax];
		w++;
		ptrs = scores0;
		scores0 = scores1;
		scores1 = ptrs;
		//for(j=0;j<=max;j++)
		//	scores0[j] = scores1[j];
	}

	//----------------------------------------------
	// the rest of the letters
	max = allowed_errors*2;
	iz = 0;
	w = allowed_errors-1;
	for (i=allowed_errors_plus_one;i<redlen;i++)
	{
		w++;
		iz++;
		scores1[0] = (sread[i]!=sref[iz]);
		if (scores0[1] < scores0[0])
			scores1[0] = scores0[1]+1;
		else scores1[0] += scores0[0];

		r_index = iz;
		//i_minus_one = i-1;
		for (j=1;j < max;j++)
		{
			l++;
			r_index++;
			scores1[j] = (sread[i]!=sref[r_index]);
			int d = scores0[j];
			int t = scores0[j+1];
			int l = scores1[j-1];
			if (t < d)
			{
				if (t <= l)
					scores1[j] = t+1;
				else scores1[j] = l+1;
			} else if (l < d)
				scores1[j] = l+1;
			else scores1[j] += d;
		}

		scores1[max] = (sread[i]!=sref[iz+max]);
		if (scores1[max-1] < scores0[max])
			scores1[max] = scores1[max-1]+1;
		else scores1[max] += scores0[max];

		//for (j=0;j<=max;j++)
		//	scores0[j] = scores1[j];

		ptrs = scores0;
		scores0 = scores1;
		scores1 = ptrs;
		//-------------------------------------------------
		// exit early routine
		//temp = max+1;
		//for (n=0;n<temp;n++)
		//	if (scores[i][n] < allowed_errors_plus_one)
		//		break;
		//if (n == max+1)
		//	return(allowed_errors_plus_one);
	}
	mi = redlen-1;
	l = scores0[0];
	temp = max+1;
	for(i=1;i<temp;i++)
		if (l > scores0[i])
			l = scores0[i];
	return l;
}

int SingleEndIndelVerifier::BandedAlign(int allowed_errors, char * sref, char * sread, int reflen, int redlen, int z, stringstream& finalcigar, unsigned& indel, unsigned& OFFSET)
{

	char scores[400][25];
	indel=0;

		for (int i=0;i<=allowed_errors;i++)
		{
			scores[0][i] = (sread[0]==sref[i]?0:1);
////			cout << (int)scores[0][i];
		}
////		cout << endl;
		//----------------------------------------------
		// letters within the allowed errors
		for (int i=1;i<allowed_errors+1;i++)
		{
			int w=i-1;

			scores[i][0] = (sread[i]==sref[0]?0:1);
			//cout << (int)scores[i][0];
			if (scores[w][0] >= i)
				scores[i][0] += i;
			else scores[i][0] = scores[w][0]+1;

////			cout << (int)scores[i][0];

			int max = allowed_errors+i;
			for (int j=1;j < max;j++)
			{
				scores[i][j] = (sread[i]==sref[j]?0:1);
				//cout << (int)scores[i][j];
//				int score = scores[i][j];
				int d = scores[w][j-1];
				int t = scores[w][j];
				int l = scores[i][j-1];
				if (t < d)
				{
					if (t <= l)
						scores[i][j] = t+1;
					else scores[i][j] = l+1;
				} else if (l < d)
					scores[i][j] = l+1;
				else scores[i][j] += d;
				//cout << " [s=" << score <<  ",t=" << t << ",l=" << l << ",d=" << d << ",w=" << w << ",j=" << j << "]";
////				cout << (int)scores[i][j];
				//cout << "[" << i << "," << j << "]";
			}
			int tmax = max-1;
			scores[i][max] = (sread[i]==sref[max]?0:1);
			//cout << (int)scores[i][max] << endl;
			if (scores[i][tmax] < scores[w][tmax])
				scores[i][max] = scores[i][tmax]+1;
			else scores[i][max] += scores[w][tmax];
////			cout << (int)scores[i][max] << endl;
		}

		//----------------------------------------------
		// the rest of the letters
		int max = allowed_errors*2;
		for (int i=allowed_errors+1;i<redlen;i++)
		{
			int w = i-1;

			scores[i][0] = (sread[i]==sref[i-allowed_errors]?0:1);
////			for (int z=0;z<i-allowed_errors;z++)
////				cout << " ";
			//cout << (int)scores[i][0];
			if (scores[w][1] < scores[w][0])
				scores[i][0] = scores[w][1]+1;
			else scores[i][0] += scores[w][0];

////			cout << (int)scores[i][0];

			for (int j=1;j < max;j++)
			{
				scores[i][j] = (sread[i]==sref[(i-allowed_errors)+j]?0:1);
//				int score  = scores[i][j];
				//cout << (int)scores[i][j];
				int d = scores[w][j];
				int t = scores[w][j+1];
				int l = scores[i][j-1];
				if (t < d)
				{
					if (t <= l)
						scores[i][j] = t+1;
					else scores[i][j] = l+1;
				} else if (l < d)
					scores[i][j] = l+1;
				else scores[i][j] += d;
				//cout << " [s=" << score <<  ",t=" << t << ",l=" << l << ",d=" << d << ",w=" << w << ",j=" << j << "]";
////				cout << (int)scores[i][j];
				//cout << "[" << i << "," << j << "]";
			}

			scores[i][max] = (sread[i]==sref[i-allowed_errors+max]?0:1);
			//cout << (int)scores[i][max] << endl;
			if (scores[i][max-1] < scores[w][max])
				scores[i][max] = scores[i][max-1]+1;
			else scores[i][max] += scores[w][max];
////			cout << (int)scores[i][max] << endl;
			//-------------------------------------------------
			// exit early routine
			int n;
			for (n=0;n<=max;n++)
				if (scores[i][n] <= allowed_errors)
					break;
			if (n == max+1)
				return(allowed_errors+1);
		}
		int col=0;
		int mi = redlen-1;
		int l = scores[mi][0];
		for(int i=1;i<=max;i++)
			if (l > scores[mi][i]){
				l = scores[mi][i];
				col=i;
			}

		// trace-back step for cigar
/*
	if(l <= allowed_errors){
		string cigar;
		if (l == 0) {
			finalcigar <<redlen << "M";
			return l;
		}
		int flag=0;
		int inserted = 0;
		int deleted = 0;
		int matched = 0;
		int curval=l;
		int currow=mi;
		int curcol= currow - allowed_errors + col;

		cigar="";
		if (flag == 0) {
			while ( currow >= 0 && curcol >=0){
				if(sread[currow]==sref[curcol] ){
					++matched;
					currow--;
					curcol--;
					if(inserted>0){
						stringstream tmpcigar;
						tmpcigar << inserted<< "I" ;
						indel += inserted;
						cigar.insert(0,tmpcigar.str());
						inserted=0;
					}
					if(deleted>0){
						stringstream tmpcigar;
						tmpcigar << deleted<< "D" ;
						indel += deleted;
						cigar.insert(0,tmpcigar.str());
						deleted=0;
					}
				}
				else if(scores[currow][col]-scores[currow-1][col]==1){
					++matched;
					currow--;
					curcol--;
					if(inserted>0){
						stringstream tmpcigar;
						tmpcigar << inserted<< "I" ;
						indel += inserted;
						cigar.insert(0,tmpcigar.str());
						inserted=0;
					}
					if(deleted>0){
						stringstream tmpcigar;
						tmpcigar << deleted<< "D" ;
						indel += deleted;
						cigar.insert(0,tmpcigar.str());
						deleted=0;
					}
				}
				else if(scores[currow][col]-scores[currow][col-1]==1){ // inserted
					curcol--;
					col--;
					curval--;
					if(matched >0){
						stringstream tmpcigar;
						tmpcigar << matched<< "M"  ;
						cigar.insert(0,tmpcigar.str());
						matched=0;
					}
					++inserted;
				}
				else {//if(scores[currow][col]-scores[currow-1][col+1]==1) // deleted
					currow--;
					col++;
					if(matched >0){
						stringstream tmpcigar;
						tmpcigar << matched<< "M"  ;
						cigar.insert(0,tmpcigar.str());
						matched=0;
					}
					++deleted;
				}
			}
			if(matched!=0){
				stringstream tmpcigar;
				tmpcigar << matched<< "M"  ;
				cigar.insert(0,tmpcigar.str());
			}
		}
		finalcigar << cigar;
	}
*/
	if(l <= allowed_errors){
		string cigar;
		if (l == 0) {
			finalcigar << redlen << "M";
			return l;
		}
		int flag=0;
		int inserted = 0;
		int deleted = 0;
		int matched = 0;
		int curval=l;
		int currow=mi;
		int curcol= currow - allowed_errors + col;

		cigar="";
		int prcol = col; //the column in currow - 1
		int maxcol = 2*allowed_errors;

		//int derr = 0;

		if (flag == 0) {
			while ( currow >= 0 && curcol >=0){
				if(currow < allowed_errors){ col--; maxcol--; }

				if(currow <= allowed_errors) prcol = col - 1;
				else prcol = col;

				if(sread[currow]==sref[curcol] ){
					++matched;
					currow--;
					curcol--;
					if(inserted>0){
						stringstream tmpcigar;
						tmpcigar << inserted<< "D";//"I" ;
						indel += inserted;
						cigar.insert(0,tmpcigar.str());
						inserted=0;
					}
					if(deleted>0){
						stringstream tmpcigar;
						tmpcigar << deleted<< "I";//"D" ;
						indel += deleted;
						cigar.insert(0,tmpcigar.str());
						deleted=0;
					}
				}
				else if(currow != 0 && prcol >= 0 && scores[currow][col]-scores[currow-1][prcol]==1){ //substituted
					++matched; //derr++;
					currow--;
					curcol--;
					if(inserted>0){
						stringstream tmpcigar;
						tmpcigar << inserted<< "D";//"I" ;
						indel += inserted;
						cigar.insert(0,tmpcigar.str());
						inserted=0;
					}
					if(deleted>0){
						stringstream tmpcigar;
						tmpcigar << deleted<< "I";//"D" ;
						indel += deleted;
						cigar.insert(0,tmpcigar.str());
						deleted=0;
					}
				}
				else if(col > 0 && scores[currow][col]-scores[currow][col-1]==1){ // inserted
					curcol--;
					col--;

					// Because currow does not decrease
					// Note that both vars decrease by 1 at the begining of the while loop,
					if(currow < allowed_errors){ col++; maxcol++; }

					curval--;
					if(matched >0){
						stringstream tmpcigar;
						tmpcigar << matched<< "M"  ;
						cigar.insert(0,tmpcigar.str());
						matched=0;
					}
					++inserted; //derr++;
				}
				else if(currow != 0 && prcol < maxcol && scores[currow][col]-scores[currow-1][prcol+1]==1){ // deleted
					currow--;
					col++;
					if(matched >0){
						stringstream tmpcigar;
						tmpcigar << matched<< "M"  ;
						cigar.insert(0,tmpcigar.str());
						matched=0;
					}
					++deleted; //derr++;
				}
				else{ // default - substituted 
					matched++; //derr++;
					currow--;
					curcol--;
					if(inserted>0){
						stringstream tmpcigar;
						tmpcigar << inserted<< "D";//"I" ;
						indel += inserted;
						cigar.insert(0,tmpcigar.str());
						inserted=0;
					}
					if(deleted>0){
						stringstream tmpcigar;
						tmpcigar << deleted<< "I";//"D" ;
						indel += deleted;
						cigar.insert(0,tmpcigar.str());
						deleted=0;
					}
				}
			}
			if(matched!=0){
				stringstream tmpcigar;
				tmpcigar << matched<< "M"  ;
				cigar.insert(0,tmpcigar.str());
			}
			if(currow >= 0){
				stringstream tmpcigar;
				tmpcigar << deleted + currow + 1 << "I"; // "D"
				cigar.insert(0,tmpcigar.str());
				indel += (deleted + currow + 1);

				//derr += (currow + 1);
			}
		}
		finalcigar << cigar;

		OFFSET = curcol + 1;
	}

	return l;
}

int SingleEndIndelVerifier::TEditDistance(int k,char * s1, char * s2, int l1)
{
    //-----------------------------------------------------------------------------
    // XOR original sequences and copy to 2 results rs and rs5
    // s1 = candidate
    // s2 = read
    //-----------------------------------------------------------------------------
    int hamming = 0;

    for (int i=0; i<l1; i++)
        if (s1[i]^s2[i])
            hamming++;

    if (hamming <= k)
        return hamming;

    if (s2[0]^s1[0] && s2[0]^s1[1] && s2[0]^s1[2])
    {
        int last = l1-1;
        for(int n=last-2;n<last+3;n++)
            if (!(s2[last]^s1[n]))
                return k+1;
        return -1;
    }
    return k+1;
}
unsigned SingleEndIndelVerifier::getBitVectorEditDistance(const char* candi, unsigned& refMatchLen,
		unsigned maxLen, const string& query, unsigned remainingDist ) {

	unsigned qlen = query.length();
	memcpy(readCopy, query.data(), qlen);
	readCopy[qlen] = '\0';
	memcpy(candiCopy, candi, maxLen);
	candiCopy[maxLen] = '\0';
	return BandedAlign((int)remainingDist, candiCopy, readCopy, maxLen, qlen,(int)remainingDist);
}

unsigned SingleEndIndelVerifier::getBitVectorEditDistance(const char* candi, unsigned& refMatchLen,
		unsigned maxLen, const string& query, unsigned remainingDist, stringstream& cigar, unsigned& indel, unsigned& OFFSET) {

	unsigned qlen = query.length();
	memcpy(readCopy, query.data(), qlen);
	readCopy[qlen] = '\0';
	memcpy(candiCopy, candi, maxLen);
	candiCopy[maxLen] = '\0';
	return BandedAlign((int)remainingDist, candiCopy, readCopy, maxLen, qlen,(int)remainingDist,cigar,indel, OFFSET);
}
unsigned SingleEndIndelVerifier::getCigar(const char* candi, const string& query,
		stringstream& cigar, unsigned refMatchLen, unsigned distance) {

	unsigned countIndel = 0;
	if (distance == 0) {
		cigar << query.length() << "M";
		return countIndel;
	}
	int flag = 0;
	unsigned qlen = query.length();
	typedef seqan::String<seqan::Dna5> TSequence;
	memcpy(readCopy, query.data(), qlen);
	readCopy[qlen] = '\0';
	cout << refMatchLen << endl;
	memcpy(candiCopy, candi, refMatchLen);
	candiCopy[refMatchLen] = '\0';


	seqan::Score<int> score(0, -1, -1, -1);
	seqan::Align<TSequence, seqan::ArrayGaps> align;
	seqan::resize(seqan::rows(align), 2);
	seqan::assignSource(seqan::row(align, 0), candiCopy);
	seqan::assignSource(seqan::row(align, 1), readCopy);


	if (refMatchLen == qlen) {
		if (distance == 1) {
			flag = 1;
			cigar << query.length() << "M";
		} else {
			SingleEndHammingVerifier hammingVerifier(indexer, distance);
			unsigned hammingDist = hammingVerifier.getDistance(candi, refMatchLen, query);
			if (hammingDist == distance) {
				flag = 1;
				cigar << qlen << "M";
			}
		}

	}
	if (flag == 0) {

		seqan::globalAlignment(align, score);
		unsigned viewPosReadLast = qlen;
		if (qlen < refMatchLen)
			viewPosReadLast = refMatchLen;
		typedef Align<TSequence, ArrayGaps> TAlign;
		typedef Row<TAlign>::Type TRow;
		typedef Iterator<TRow, Rooted>::Type TAlignIterator;
		typedef Source<TAlign>::Type TSource;
		typedef Iterator<TSource, Rooted>::Type TStringIterator;
		TAlignIterator ali_it0 = iter(row(align, 0), 0);
		TAlignIterator ali_it1 = iter(row(align, 1), 0);
		TAlignIterator ali_it0_stop = iter(row(align, 0), viewPosReadLast);
		TAlignIterator ali_it1_stop = iter(row(align, 1), viewPosReadLast);
		int readPos = 0;
		TStringIterator readBase = begin(source(row(align, 0)));
		bool first = true;
		while (ali_it0 != ali_it0_stop && ali_it1 != ali_it1_stop) {
			int inserted = 0;
			int deleted = 0;
			int matched = 0;
			while (ali_it0 != ali_it0_stop && ali_it1 != ali_it1_stop && !isGap(ali_it0)
					&& !isGap(ali_it1)) {
				++readPos;
				if (*ali_it1 != *ali_it0 && first) {
					first = false;
				}
				++readBase;
				++ali_it0;
				++ali_it1;
				++matched;
			}
			if (matched > 0) {
				cigar << matched << "M";
			}
			while (ali_it0 != ali_it0_stop && isGap(ali_it0)) {
				++ali_it0;
				++ali_it1;
				++deleted;
			}
			if (deleted > 0) {
				cigar << deleted << "D";
				countIndel += deleted;
			}
			while (isGap(ali_it1) && ali_it1 != ali_it1_stop) {
				++ali_it0;
				++ali_it1;
				++readPos;
				if (first) {
					first = false;
				}
				++readBase;
				++inserted;
			}
			if (inserted > 0) {
				cigar << inserted << "I";
				countIndel += inserted;
			}

		}
	}
	return countIndel;
}

unsigned SingleEndIndelVerifier::getDistance(const char* candi, unsigned candiLen,
		const string& read, int& offset, unsigned& refMatchLen) {
	unsigned distance = getBitVectorEditDistance(candi, refMatchLen, candiLen + distThresh, read,
			distThresh);
	unsigned startPosMaxShift = 1;
	if ((distance > distThresh + startPosMaxShift) || (distance <= distThresh)) {
		return distance;
	}
	for (unsigned i = 1; i <= startPosMaxShift; i++) {
		distance = getBitVectorEditDistance(candi + i, refMatchLen, candiLen + distThresh, read,
				distThresh);
		if (distance <= distThresh) {
			offset = i;
			return distance;
		}
		distance = getBitVectorEditDistance(candi - i, refMatchLen, candiLen + distThresh, read,
				distThresh);
		if (distance <= distThresh) {
			offset = -i;
			return distance;
		}
	}
	return distance;
}

unsigned SingleEndIndelVerifier::getDistance(const char* candi, unsigned candiLen, const string& query) {
	int unusedOffset;
	unsigned unusedRefMatchLen;
	return getDistance(candi, candiLen, query, unusedOffset, unusedRefMatchLen);
}

unsigned SingleEndIndelVerifier::getDistanceLowerBound(GramCountArray& readCounts,
		unsigned refPos, unsigned refLen) {
	GramCountArray candidateCount;
	fillGramCountArray((char*) refSeqPtr + refPos, refLen, candidateCount);
	unsigned count1 = 0;
	unsigned count2 = 0;
	for (unsigned i = 0; i < GRAM_COUNT_ARRAY_LEN; i++) {
		if (readCounts.count[i] > candidateCount.count[i]) {
			count1 += readCounts.count[i] - candidateCount.count[i];
		} else {
			count2 += candidateCount.count[i] - readCounts.count[i];
		}
	}
	unsigned lowerBound = (max(count1, count2) + COUNT_FILTER_GRAM_LEN - 1) / COUNT_FILTER_GRAM_LEN;
	return lowerBound;
}

void SingleEndIndelVerifier::fillGramCountArray(const char* candi, unsigned candiLen,
		GramCountArray& counts) {
	unsigned mask = (numeric_limits<unsigned>::max() >> (32 - 2 * COUNT_FILTER_GRAM_LEN));
	unsigned hash = 0;
	unsigned i = 0;
	unsigned j = 0;
	while (j < i + COUNT_FILTER_GRAM_LEN && j < candiLen) {
		switch (candi[j]) {
		case BASE_A:
			hash = hash << 2;
			break;
		case BASE_C:
			hash = (hash << 2) + 1;
			break;
		case BASE_G:
			hash = (hash << 2) + 2;
			break;
		case BASE_T:
			hash = (hash << 2) + 3;
			break;
		default:
			hash = 0;
			i = j + 1;
			break;
		}
		j++;
	}
	if (j > candiLen) {
		return;
	}
	i = j;
	counts.count[hash]++;
	while (i < candiLen) {
		switch (candi[i]) {
		case BASE_A:
			hash = (hash << 2) & mask;
			break;
		case BASE_C:
			hash = ((hash << 2) + 1) & mask;
			break;
		case BASE_G:
			hash = ((hash << 2) + 2) & mask;
			break;
		case BASE_T:
			hash = ((hash << 2) + 3) & mask;
			break;
		default:
			i++;
			hash = 0;
			j = i;
			while (j < i + COUNT_FILTER_GRAM_LEN && j < candiLen) {
				switch (candi[j]) {
				case BASE_A:
					hash = hash << 2;
					break;
				case BASE_C:
					hash = (hash << 2) + 1;
					break;
				case BASE_G:
					hash = (hash << 2) + 2;
					break;
				case BASE_T:
					hash = (hash << 2) + 3;
					break;
				default:
					hash = 0;
					i = j + 1;
					break;
				}
				j++;
			}
			if (j > candiLen) {
				return;
			} else {
				i = j - 1;
			}
			break;
		}
		counts.count[hash]++;
		i++;
	}
}

bool cmp_readmap(const ReadMapping& r1, const ReadMapping& r2)
{
	return r1.pos < r2.pos;
}
void SingleEndHeuristicIndelVerifier::verifyCandidates(const string& read,
		//vector<unsigned>& candidates, vector<ReadMapping>& results) {
		vector<unsigned>& candidates, vector<ReadMapping>& res) {
	unsigned bestAnswerSoFar;
	unsigned readlen = read.size();
	GramCountArray readGramCounts;
	fillGramCountArray(read.data(), read.length(), readGramCounts);
	unsigned* candi = candidates.data();
	unsigned* candiEnd = candidates.data() + candidates.size();

	unsigned OFFSET = 0;

	vector<ReadMapping> results;

	while (candi != candiEnd) {
		unsigned currentCandidate = *candi++;
//		unsigned resultForThisCandidate = 0;
		bestAnswerSoFar = distThresh + 1;
		__builtin_prefetch((const void*) (refSeqPtr + (*(candi + CACHE_PREFETCH_STEP))), 0, 0);
		unsigned maxLength = readlen + distThresh;
		unsigned lowerBound = getDistanceLowerBound(readGramCounts, currentCandidate, readlen);
		unsigned dist = -1;
		unsigned refMatchLen = -1;
		if (lowerBound <= distThresh) {
			stringstream* cigar = NULL;
			if(genCigar){
				cigar = new stringstream();
				unsigned indel=-1;
				dist = getBitVectorEditDistance(refSeqPtr + currentCandidate, refMatchLen, maxLength,
						read, distThresh,*cigar,indel, OFFSET);

				if (dist <= distThresh) {
					vector<unsigned>::iterator it = upper_bound(indexer->chromosomeLengths.begin(),
							indexer->chromosomeLengths.end(), currentCandidate);
					if (currentCandidate + read.size() <= *it) {
						if(indel >0){
							tempIndel[indel]++;
						}
						else{
							tempMisMatch[dist]++;
						}
						results.push_back(ReadMapping(currentCandidate + OFFSET, dist, cigar));
					}
					else
						delete cigar;
				}
				else
					delete cigar;
			}
			else{
				dist = getBitVectorEditDistance(refSeqPtr + currentCandidate, refMatchLen, maxLength,read, distThresh);
				if (dist <= distThresh) {
					vector<unsigned>::iterator it = upper_bound(indexer->chromosomeLengths.begin(),
							indexer->chromosomeLengths.end(), currentCandidate);
					if (currentCandidate + read.size() <= *it) {
						tempMisMatch[dist]++;
						results.push_back(ReadMapping(currentCandidate, dist, cigar));
					}
					continue;
				}
			}
		}
	}

	std::sort(results.begin(), results.end(), cmp_readmap);

	for(int i = 0; i < results.size(); i++){
		if(i != 0 && results[i].pos == results[i-1].pos && results[i].distance == results[i-1].distance) continue;
		res.push_back(results[i]);
	}
}

void SingleEndHeuristicIndelVerifier::verifyCandidates(unsigned& startPos, unsigned& k,
		const string& read, vector<unsigned>& candidates, vector<ReadMapping>& results) {
	unsigned bestDistance;
	GramCountArray queryGramCounts;
	fillGramCountArray(read.data(), read.length(), queryGramCounts);
	unsigned readlen = read.size();
	unsigned* candi = candidates.data() + startPos;
	unsigned* candiEnd = candidates.data() + candidates.size();
	while (candi != candiEnd) {
		unsigned currentCandidate = *candi++;
		unsigned resultForThisCandidate = 0;
		__builtin_prefetch((const void*) (refSeqPtr + (*(candi + CACHE_PREFETCH_STEP))), 0, 0);
		bestDistance = distThresh + 1;
		unsigned refMatchLen = -1;
		unsigned lowerBound = getDistanceLowerBound(queryGramCounts, currentCandidate, readlen);
		if (lowerBound <= distThresh) {
			unsigned maxLength = readlen + distThresh;
			unsigned distance = getBitVectorEditDistance(refSeqPtr + currentCandidate, refMatchLen,
					maxLength, read, distThresh);
			if (distance <= distThresh && distance > 0) {
				vector<unsigned>::iterator it = upper_bound(indexer->chromosomeLengths.begin(),
						indexer->chromosomeLengths.end(), currentCandidate);
				if (currentCandidate + readlen <= *it) {
					stringstream* cigar = NULL;
					if (genCigar) {
						cigar = new stringstream();
						unsigned indel = getCigar(refSeqPtr + currentCandidate, read, *cigar,
								refMatchLen, distance);
						if (indel > 0) {
							tempIndel[indel]++;
						} else {
							tempMisMatch[distance]++;
						}
					} else {
						tempMisMatch[distance]++;
					}
					results.push_back(ReadMapping(currentCandidate, distance, cigar));
					k--;
					if (k == 0) {
						return;
					}
					continue;
				}
			}
		}
		unsigned stRange = 1;
		for (unsigned startLocRange = 1; startLocRange <= stRange; startLocRange++) {
			if (lowerBound > (startLocRange + distThresh)) {
				continue;
			}
			unsigned loc = currentCandidate - startLocRange;
			unsigned maxLength =
					((loc) + readlen + distThresh) < refSeq.length() ?
							(readlen + distThresh) : (refSeq.length() - (loc));
			lowerBound = getDistanceLowerBound(queryGramCounts, loc, readlen);
			if (lowerBound <= distThresh) {
				unsigned distance = getBitVectorEditDistance(refSeqPtr + loc, refMatchLen,
						maxLength, read, distThresh);
				if (distance <= distThresh && distance > 0) {
					bestDistance = distance;
					resultForThisCandidate = loc;
					break;
				}
			}
			loc = currentCandidate + startLocRange;
			maxLength =
					((loc) + readlen + distThresh) < refSeq.length() ?
							(readlen + distThresh) : (refSeq.length() - (loc));
			lowerBound = getDistanceLowerBound(queryGramCounts, loc, readlen);
			if (lowerBound <= distThresh) {
				unsigned distance = getBitVectorEditDistance(refSeqPtr + loc, refMatchLen,
						maxLength, read, distThresh);
				if (distance <= distThresh && distance > 0) {
					bestDistance = distance;
					resultForThisCandidate = loc;
					break;
				}
			}
		}
		if (bestDistance <= distThresh) {
			vector<unsigned>::iterator it = upper_bound(indexer->chromosomeLengths.begin(),
					indexer->chromosomeLengths.end(), resultForThisCandidate);
			if (resultForThisCandidate + read.size() <= *it) {
				stringstream* cigar = NULL;
				if (genCigar) {
					cigar = new stringstream();
					unsigned indel = getCigar(refSeqPtr + resultForThisCandidate, read, *cigar,
							refMatchLen, bestDistance);
					if (indel > 0) {
						tempIndel[indel]++;
					} else {
						tempMisMatch[bestDistance]++;
					}
				} else {
					tempMisMatch[bestDistance]++;
				}
				results.push_back(ReadMapping(resultForThisCandidate, bestDistance, cigar));
				k--;
				if (k == 0) {
					return;
				}
				continue;
			}
		}
	}
}

